import ubelt as ub
import os
import math
import pandas as pd
import pygtrie
import kwarray
import wrapt
from kwutil import slugify_ext
from packaging.version import parse as Version
from geowatch.utils.util_stringalgo import shortest_unique_suffixes

PANDAS_GE_2_0_0 = Version(pd.__version__) >= Version('2.0.0')


class DataFrame(pd.DataFrame):
    """
    Extension of pandas dataframes with quality-of-life improvements.

    Refernces:
        .. [SO22155951] https://stackoverflow.com/questions/22155951/how-can-i-subclass-a-pandas-dataframe

    Example:
        from geowatch.utils.util_pandas import *  # NOQA
        from geowatch.utils import util_pandas
        df = util_pandas.DataFrame.random()
    """
    # _metadata = ['added_property']
    # added_property = 1  # This will be passed to copies

    @property
    def _constructor(self):
        return DataFrame

    @classmethod
    def random(cls, rows=10, columns='abcde', rng=None):
        """
        Create a random data frame for testing.

        rows=10
        columns='abcde'
        rng = None
        cls = util_pandas.DataFrame
        """
        import kwarray
        rng = kwarray.ensure_rng(rng)

        def coerce_index(data):
            if isinstance(data, int):
                return list(range(data))
            else:
                return list(data)
        columns = coerce_index(columns)
        index = coerce_index(rows)
        random_data = [{c: rng.rand() for c in columns} for r in index]
        self = cls(random_data, index=index, columns=columns)
        return self

    @classmethod
    def coerce(cls, data):
        """
        Ensures that the input is an instance of our extended DataFrame.

        Pandas is generally good about input coercion via its normal
        constructors, the purpose of this classmethod is to quickly ensure that
        a DataFrame has all of the extended methods defined by this class
        without incurring a copy. In this sense it is more similar to
        :func:numpy.asarray`.

        Args:
            data (DataFrame | ndarray | Iterable | dict):
                generally another dataframe, otherwise normal inputs that would
                be given to the regular pandas dataframe constructor

        Returns:
            DataFrame:

        Example:
            >>> # xdoctest: +REQUIRES(--benchmark)
            >>> # This example demonstrates the speed difference between
            >>> # recasting as a DataFrame versus using coerce
            >>> from geowatch.utils.util_pandas import DataFrame
            >>> data = DataFrame.random(rows=10_000)
            >>> import timerit
            >>> ti = timerit.Timerit(100, bestof=10, verbose=2)
            >>> for timer in ti.reset('constructor'):
            >>>     with timer:
            >>>         DataFrame(data)
            >>> for timer in ti.reset('coerce'):
            >>>     with timer:
            >>>         DataFrame.coerce(data)
            >>> # xdoctest: +IGNORE_WANT
            Timed constructor for: 100 loops, best of 10
                time per loop: best=2.594 µs, mean=2.783 ± 0.1 µs
            Timed coerce for: 100 loops, best of 10
                time per loop: best=246.000 ns, mean=283.000 ± 32.4 ns
        """
        if isinstance(data, cls):
            return data
        else:
            return cls(data)

    def safe_drop(self, labels, axis=0):
        """
        Like :func:`self.drop`, but does not error if the specified labels do
        not exist.

        Args:
            df (pd.DataFrame): df
            labels (List): ...
            axis (int): todo

        Example:
            >>> from geowatch.utils.util_pandas import *  # NOQA
            >>> import numpy as np
            >>> self = DataFrame({k: np.random.rand(10) for k in 'abcde'})
            >>> self.safe_drop(list('bdf'), axis=1)
        """
        existing = self.axes[axis]
        labels = existing.intersection(labels)
        return self.drop(labels, axis=axis)

    def reorder(self, head=None, tail=None, axis=0, missing='error',
                fill_value=float('nan'), **kwargs):
        """
        Change the order of the row or column index. Unspecified labels will
        keep their existing order after the specified labels.

        Args:
            head (List | None):
                The order of the labels to put at the start of the re-indexed
                data frame. Unspecified labels keep their relative order and
                are placed after specified these "head" labels.

            tail (List | None):
                The order of the labels to put at the end of the re-indexed
                data frame. Unspecified labels keep their relative order and
                are placed after before these "tail" labels.

            axis (int):
                The axis 0 for rows, 1 for columns to reorder.

            missing (str):
                Policy to handle specified labels that do not exist in the
                specified axies. Can be either "error", "drop", or "fill".
                If "drop", then drop any specified labels that do not exist.
                If "error", then raise an error non-existing labels are given.
                If "fill", then fill in values for labels that do not exist.

            fill_value (Any):
                fill value to use when missing is "fill".

        Returns:
            Self - DataFrame with modified indexes

        Example:
            >>> from geowatch.utils import util_pandas
            >>> self = util_pandas.DataFrame.random(rows=5, columns=['a', 'b', 'c', 'd', 'e', 'f'])
            >>> new = self.reorder(['b', 'c'], axis=1)
            >>> assert list(new.columns) == ['b', 'c', 'a', 'd', 'e', 'f']
            >>> # Set the order of the first and last of the columns
            >>> new = self.reorder(head=['b', 'c'], tail=['e', 'd'], axis=1)
            >>> assert list(new.columns) == ['b', 'c', 'a', 'f', 'e', 'd']
            >>> # Test reordering the rows
            >>> new = self.reorder([1, 0], axis=0)
            >>> assert list(new.index) == [1, 0, 2, 3, 4]
            >>> # Test reordering with a non-existent column
            >>> new = self.reorder(['q'], axis=1, missing='drop')
            >>> assert list(new.columns) == ['a', 'b', 'c', 'd', 'e', 'f']
            >>> new = self.reorder(['q'], axis=1, missing='fill')
            >>> assert list(new.columns) == ['q', 'a', 'b', 'c', 'd', 'e', 'f']
            >>> import pytest
            >>> with pytest.raises(ValueError):
            >>>     self.reorder(['q'], axis=1, missing='error')
            >>> # Should error if column is given in both head and tail
            >>> with pytest.raises(ValueError):
            >>>     self.reorder(['c'], ['c'], axis=1, missing='error')
        """
        if 'intersect' in kwargs:
            raise Exception('The intersect argument was removed. Set missing=drop')
        if kwargs:
            raise ValueError(f'got unknown kwargs: {list(kwargs.keys())}')

        existing = self.axes[axis]
        if head is None:
            head = []
        if tail is None:
            tail = []
        head_set = set(head)
        tail_set = set(tail)
        duplicate_labels = head_set & tail_set
        if duplicate_labels:
            raise ValueError(
                'Cannot specify the same label in both the head and tail.'
                f'Duplicate labels: {duplicate_labels}')
        if missing == 'drop':
            orig_order = ub.oset(list(existing))
            resolved_head = ub.oset(head) & orig_order
            resolved_tail = ub.oset(tail) & orig_order
        elif missing == 'error':
            requested = (head_set | tail_set)
            unknown = requested - set(existing)
            if unknown:
                raise ValueError(
                    f"Requested labels that don't exist unknown={unknown}. "
                    "Specify intersect=True to ignore them.")
            resolved_head = head
            resolved_tail = tail
        elif missing == 'fill':
            resolved_head = head
            resolved_tail = tail
        else:
            raise KeyError(missing)
        remain = existing.difference(resolved_head).difference(resolved_tail)
        new_labels = list(resolved_head) + list(remain) + list(resolved_tail)
        return self.reindex(labels=new_labels, axis=axis,
                            fill_value=fill_value)

    def _orig_groupby(self, by=None, **kwargs):
        return super().groupby(by=by, **kwargs)

    def groupby(self, by=None, **kwargs):
        """
        Fixed groupby behavior so length-one arguments are handled correctly

        Args:
            df (DataFrame):
            ** kwargs: groupby kwargs

        Example:
            >>> from geowatch.utils import util_pandas
            >>> df = util_pandas.DataFrame({
            >>>     'Animal': ['Falcon', 'Falcon', 'Parrot', 'Parrot'],
            >>>     'Color': ['Blue', 'Blue', 'Blue', 'Yellow'],
            >>>     'Max Speed': [380., 370., 24., 26.]
            >>>     })
            >>> new1 = dict(list(df.groupby(['Animal', 'Color'])))
            >>> new2 = dict(list(df.groupby(['Animal'])))
            >>> new3 = dict(list(df.groupby('Animal')))
            >>> assert sorted(new1.keys())[0] == ('Falcon', 'Blue')
            >>> assert sorted(new3.keys())[0] == 'Falcon'
            >>> # This is the case that is fixed.
            >>> assert sorted(new2.keys())[0] == ('Falcon',)
        """
        groups = super().groupby(by=by, **kwargs)
        fixed_groups = _fix_groupby(groups)
        return fixed_groups

    def match_columns(self, pat, hint='glob'):
        """
        Find matching columns in O(N)
        """
        from kwutil import util_pattern
        pat = util_pattern.Pattern.coerce(pat, hint=hint)
        found = [c for c in self.columns if pat.match(c)]
        return found

    def search_columns(self, pat, hint='glob'):
        """
        Find matching columns in O(N)
        """
        from kwutil import util_pattern
        pat = util_pattern.Pattern.coerce(pat, hint=hint)
        found = [c for c in self.columns if pat.search(c)]
        return found

    def varied_values(self, **kwargs):
        """
        Kwargs:
            min_variations=0, max_variations=None, default=ub.NoParam,
            dropna=False, on_error='raise'

        SeeAlso:
            :func:`geowatch.utils.result_analysis.varied_values`
        """
        from geowatch.utils.result_analysis import varied_values
        varied = varied_values(self, **kwargs)
        return varied

    def varied_value_counts(self, **kwargs):
        """
        Kwargs:
            min_variations=0, max_variations=None, default=ub.NoParam,
            dropna=False, on_error='raise'

        SeeAlso:
            :func:`geowatch.utils.result_analysis.varied_value_counts`
        """
        from geowatch.utils.result_analysis import varied_value_counts
        varied = varied_value_counts(self, **kwargs)
        return varied

    def shorten_columns(self, return_mapping=False, min_length=0):
        """
        Shorten column names by separating unique suffixes based on the "."
        separator.

        Args:
            return_mapping (bool):
                if True, returns the

            min_length (int):
                minimum size of the new column names in terms of parts.

        Returns:
            DataFrame | Tuple[DataFrame, Dict[str, str]]:
                Either the new data frame with shortened column names or that
                data frame and the mapping from old column names to new column
                names.

        Example:
            >>> from geowatch.utils.util_pandas import DataFrame
            >>> # If all suffixes are unique, then they are used.
            >>> self = DataFrame.random(columns=['id', 'params.metrics.f1', 'params.metrics.acc', 'params.fit.model.lr', 'params.fit.data.seed'])
            >>> new = self.shorten_columns()
            >>> assert list(new.columns) == ['id', 'f1', 'acc', 'lr', 'seed']
            >>> # Conflicting suffixes impose limitations on what can be shortened
            >>> self = DataFrame.random(columns=['id', 'params.metrics.magic', 'params.metrics.acc', 'params.fit.model.lr', 'params.fit.data.magic'])
            >>> new = self.shorten_columns()
            >>> assert list(new.columns) == ['id', 'metrics.magic', 'metrics.acc', 'model.lr', 'data.magic']
        """
        import ubelt as ub
        old_cols = self.columns
        new_cols = shortest_unique_suffixes(old_cols, sep='.', min_length=min_length)
        mapping = ub.dzip(old_cols, new_cols)
        new = self.rename(columns=mapping)
        if return_mapping:
            return new, mapping
        else:
            return new

    def _to_dotdict(self):
        """
        Experimental, convert a a dotdict (should maybe give useful dotdict
        methods to this class?)
        """
        return DotDictDataFrame(self)

    def argextrema(self, columns, objective='maximize', k=1):
        """
        Finds the top K indexes (locs) for given columns.

        Args:
            columns (str | List[str]) : columns to find extrema of.
                If multiple are given, then secondary columns are used as
                tiebreakers.

            objective (str | List[str]) :
                Either maximize or minimize (max and min are also accepted).
                If given as a list, it specifies the criteria for each column,
                which allows for a mix of maximization and minimization.

            k : number of top entries

        Returns:
            List: indexes into subset of data that are in the top k for any of the
                requested columns.

        Example:
            >>> from geowatch.utils.util_pandas import DataFrame
            >>> # If all suffixes are unique, then they are used.
            >>> self = DataFrame.random(columns=['id', 'f1', 'loss'], rows=10)
            >>> self.loc[3, 'f1'] = 1.0
            >>> self.loc[4, 'f1'] = 1.0
            >>> self.loc[5, 'f1'] = 1.0
            >>> self.loc[3, 'loss'] = 0.2
            >>> self.loc[4, 'loss'] = 0.3
            >>> self.loc[5, 'loss'] = 0.1
            >>> columns = ['f1', 'loss']
            >>> k = 4
            >>> top_indexes = self.argextrema(columns=columns, k=k, objective=['max', 'min'])
            >>> assert len(top_indexes) == k
            >>> print(self.loc[top_indexes])
        """
        ascending = None
        def rectify_ascending(objective_str):
            if objective_str in {'max', 'maximize'}:
                ascending = False
            elif objective_str in {'min', 'minimize'}:
                ascending = True
            else:
                raise KeyError(objective)
            return ascending

        if isinstance(objective, str):
            ascending = rectify_ascending(objective)
        else:
            ascending = [rectify_ascending(o) for o in objective]

        ranked_data = self.sort_values(columns, ascending=ascending)
        if isinstance(k, float) and math.isinf(k):
            k = None
        top_locs = ranked_data.index[0:k]
        return top_locs


def pandas_reorder_columns(df, columns):
    """
    DEPRECATED: Use :func:`DataFrame.reorder` instead
    """
    remain = df.columns.difference(columns)
    return df.reindex(columns=(columns + list(remain)))


def pandas_argmaxima(data, columns, k=1):
    """
    Finds the top K indexes for given columns.

    Args:
        data : pandas data frame

        columns : columns to maximize.
            If multiple are given, then secondary columns are used as
            tiebreakers.

        k : number of top entries

    Returns:
        List: indexes into subset of data that are in the top k for any of the
            requested columns.

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> import numpy as np
        >>> import pandas as pd
        >>> data = pd.DataFrame({k: np.random.rand(10) for k in 'abcde'})
        >>> columns = ['b', 'd', 'e']
        >>> k = 1
        >>> top_indexes = pandas_argmaxima(data=data, columns=columns, k=k)
        >>> assert len(top_indexes) == k
        >>> print(data.loc[top_indexes])
    """
    ranked_data = data.sort_values(columns, ascending=False)
    if isinstance(k, float) and math.isinf(k):
        k = None
    top_locs = ranked_data.index[0:k]
    return top_locs


def pandas_suffix_columns(data, suffixes):
    """
    Return columns that end with this suffix
    """
    return [c for c in data.columns if any(c.endswith(s) for s in suffixes)]


def pandas_nan_eq(a, b):
    nan_flags1 = pd.isna(a)
    nan_flags2 = pd.isna(b)
    eq_flags = a == b
    both_nan = nan_flags1 & nan_flags2
    flags = eq_flags | both_nan
    return flags


def pandas_shorten_columns(summary_table, return_mapping=False, min_length=0):
    """
    Shorten column names

    DEPRECATED: Use :func:`DataFrame.shorten_columns` instead.

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> df = pd.DataFrame([
        >>>     {'param_hashid': 'badbeaf', 'metrics.eval.f1': 0.9, 'metrics.eval.mcc': 0.8, 'metrics.eval.acc': 0.3},
        >>>     {'param_hashid': 'decaf', 'metrics.eval.f1': 0.6, 'metrics.eval.mcc': 0.2, 'metrics.eval.acc': 0.4},
        >>>     {'param_hashid': 'feedcode', 'metrics.eval.f1': 0.5, 'metrics.eval.mcc': 0.3, 'metrics.eval.acc': 0.1},
        >>> ])
        >>> print(df.to_string(index=0))
        >>> df2 = pandas_shorten_columns(df)
        param_hashid  metrics.eval.f1  metrics.eval.mcc  metrics.eval.acc
             badbeaf              0.9               0.8               0.3
               decaf              0.6               0.2               0.4
            feedcode              0.5               0.3               0.1
        >>> print(df2.to_string(index=0))
        param_hashid  f1  mcc  acc
             badbeaf 0.9  0.8  0.3
               decaf 0.6  0.2  0.4
            feedcode 0.5  0.3  0.1

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> df = pd.DataFrame([
        >>>     {'param_hashid': 'badbeaf', 'metrics.eval.f1.mean': 0.9, 'metrics.eval.f1.std': 0.8},
        >>>     {'param_hashid': 'decaf', 'metrics.eval.f1.mean': 0.6, 'metrics.eval.f1.std': 0.2},
        >>>     {'param_hashid': 'feedcode', 'metrics.eval.f1.mean': 0.5, 'metrics.eval.f1.std': 0.3},
        >>> ])
        >>> df2 = pandas_shorten_columns(df, min_length=2)
        >>> print(df2.to_string(index=0))
        param_hashid  f1.mean  f1.std
             badbeaf      0.9     0.8
               decaf      0.6     0.2
            feedcode      0.5     0.3
    """
    import ubelt as ub
    # fixme
    old_cols = summary_table.columns
    new_cols = shortest_unique_suffixes(old_cols, sep='.', min_length=min_length)
    mapping = ub.dzip(old_cols, new_cols)
    summary_table = summary_table.rename(columns=mapping)
    if return_mapping:
        return summary_table, mapping
    else:
        return summary_table


def pandas_condense_paths(colvals):
    """
    Condense a column of paths to keep only the shortest distinguishing
    suffixes

    Args:
        colvals (pd.Series): a column containing paths to condense

    Returns:
        Tuple: the condensed series and a mapping from old to new

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> rows = [
        >>>     {'path1': '/path/to/a/file1'},
        >>>     {'path1': '/path/to/a/file2'},
        >>> ]
        >>> colvals = pd.DataFrame(rows)['path1']
        >>> pandas_condense_paths(colvals)
    """
    is_valid = ~pd.isnull(colvals)
    valid_vals = colvals[is_valid].apply(os.fspath)
    unique_valid_vals = valid_vals.unique().tolist()
    unique_short_vals = shortest_unique_suffixes(unique_valid_vals, sep='/')
    new_vals = [p.split('.')[0] for p in unique_short_vals]
    mapper = ub.dzip(unique_valid_vals, new_vals)
    condensed = colvals.apply(lambda x: mapper.get(x, x))
    return condensed, mapper


def pandas_truncate_items(data, paths=False, max_length=16):
    """
    from geowatch.utils.util_pandas import pandas_truncate_items

    Args:
        data (pd.DataFrame): data frame to truncate

    Returns:
        Tuple[pd.DataFrame, Dict[str, str]]
    """
    def truncate(x):
        if not isinstance(x, (str, os.PathLike)):
            return x
        return slugify_ext.smart_truncate(str(x), max_length=max_length, trunc_loc=0,
                                          hash_len=4, head='', tail='')
    mappings = {}
    if len(data):
        # only check the first row to see if we want to truncate the columns or
        # not
        trunc_str_cols = set()
        trunc_path_cols = set()
        for _, check_row in data.iloc[0:10].iterrows():
            for k, v in check_row.items():
                if paths:
                    # Check if probably a path or not
                    if isinstance(v, os.PathLike):
                        trunc_path_cols.add(k)
                        continue
                    elif isinstance(v, str) and '/' in v:
                        trunc_path_cols.add(k)
                        continue
                if isinstance(v, (str, os.PathLike)) and len(str(v)) > max_length:
                    trunc_str_cols.add(k)

        trunc_str_cols = list(ub.oset(data.columns) & trunc_str_cols)
        trunc_path_cols = list(ub.oset(data.columns) & trunc_path_cols)

        trunc_data = data.copy()
        trunc_data[trunc_str_cols] = trunc_data[trunc_str_cols].applymap(truncate)
        for c in trunc_data[trunc_str_cols]:
            v2 = pd.Categorical(data[c])
            data[c] = v2
            v1 = v2.map(truncate)
            mapping = list(zip(v1.categories, v2.categories))
            mappings[c] = mapping

        for c in trunc_path_cols:
            colvals = trunc_data[c]
            condensed, mapping = pandas_condense_paths(colvals)
            trunc_data[c] = condensed
            mappings[c] = mapping
    else:
        mapping = {}
        trunc_data = data
    return trunc_data, mappings


class DotDictDataFrame(pd.DataFrame):
    """
    A proof-of-concept wrapper around pandas that lets us walk down the nested
    structure a little easier.

    The API is a bit weird, and the caches are not invalidated if any column
    changes, but it does a reasonable job otherwise.

    Is there another library out there that does this?

    SeeAlso:
        DotDict

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> rows = [
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>>     {'node1.id': 1, 'node2.id': 2, 'node1.metrics.ap': 0.5, 'node2.metrics.ap': 0.8},
        >>> ]
        >>> self = DotDictDataFrame(rows)
        >>> # Test prefix lookup
        >>> assert set(self['node1'].columns) == {'node1.id', 'node1.metrics.ap'}
        >>> # Test suffix lookup
        >>> assert set(self['id'].columns) == {'node1.id', 'node2.id'}
        >>> # Test mid-node lookup
        >>> assert set(self['metrics'].columns) == {'node1.metrics.ap', 'node2.metrics.ap'}
        >>> # Test single lookup
        >>> assert set(self[['node1.id']].columns) == {'node1.id'}
        >>> # Test glob
        >>> assert set(self.find_columns('*metri*')) == {'node1.metrics.ap', 'node2.metrics.ap'}
    """

    # Not sure how safe it is to do this.
    # Consider the case of the dataframe with columns ['a.b.c', 'b.c'].
    # Asking for ['b.c'] would return both, there is no way to just get the
    # specific b.c column because its a suffix of another columns.
    # We would need to disallow the current __getitem__ behavior in order to
    # make a consistent variant of this, and perhaps thats an ok idea. Perhaps
    # we do a df.nest[<index>] localizer similar to df.loc or df.iloc and keep
    # default getitem behavior. That seems cleaner.
    #
    # @property
    # def _constructor(self):
    #     return DotDictDataFrame

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.__dict__['_trie_cache'] = {}

    def _clear_column_caches(self):
        self._trie_cache = {}

    @property
    def _column_prefix_trie(self):
        # TODO: cache the trie correctly
        if self._trie_cache.get('prefix_trie', None) is None:
            _trie_data = ub.dzip(self.columns, self.columns)
            _trie = pygtrie.StringTrie(_trie_data, separator='.')
            self._trie_cache['prefix_trie'] = _trie
        return self._trie_cache['prefix_trie']

    @property
    def _column_suffix_trie(self):
        if self._trie_cache.get('suffix_trie', None) is None:
            reversed_columns = ['.'.join(col.split('.')[::-1])
                                for col in self.columns]
            _trie_data = ub.dzip(reversed_columns, reversed_columns)
            _trie = pygtrie.StringTrie(_trie_data, separator='.')
            self._trie_cache['suffix_trie'] = _trie
        return self._trie_cache['suffix_trie']

    @property
    def _column_node_groups(self):
        if self._trie_cache.get('node_groups', None) is None:
            paths = [col.split('.') for col in self.columns]
            lut = ub.ddict(list)
            for path in paths:
                col = '.'.join(path)
                for part in path:
                    lut[part].append(col)
            self._trie_cache['node_groups'] = lut
        return self._trie_cache['node_groups']

    @property
    def nested_columns(self):
        from geowatch.utils.util_dotdict import dotkeys_to_nested
        return dotkeys_to_nested(self.columns)

    def find_column(self, col):
        result = self.query_column(col)
        if len(result) == 0:
            raise KeyError
        elif len(result) > 1:
            raise RuntimeError
        return list(result)[0]

    def query_column(self, col):
        # Might be better to do a globby sort of pattern
        parts = col.split('.')
        return ub.oset.intersection(*[self._column_node_groups[p] for p in parts])
        # try:
        #     candiates.update(self._column_prefix_trie.values(col))
        # except KeyError:
        #     ...
        # try:
        #     candiates.update(self._column_suffix_trie.values(col))
        # except KeyError:
        #     ...
        # return candiates

    def _column_graph(self):
        import networkx as nx
        graph = nx.DiGraph()
        # root = '__root__'
        # graph.add_node(root)
        for c in self.columns:
            parts = c.split('.')
            # prev_node = root
            prev_node = None
            for i in range(1, len(parts)):
                node = '.'.join(parts[:i + 1])
                graph.add_node(node, label=f'{parts[i]}')
                if prev_node is not None:
                    graph.add_edge(prev_node, node)
                prev_node = node
        nx.write_network_text(graph, with_labels=1)

    def lookup_suffix_columns(self, col):
        return self._column_suffix_trie.values(col)

    def lookup_prefix_columns(self, col):
        return self._column_prefix_trie.values(col)

    def find_columns(self, pat, hint='glob'):
        # DEPRECATE use match or search columns instead.
        from kwutil import util_pattern
        pat = util_pattern.Pattern.coerce(pat, hint=hint)
        found = [c for c in self.columns if pat.match(c)]
        return found

    def match_columns(self, pat, hint='glob'):
        from kwutil import util_pattern
        pat = util_pattern.Pattern.coerce(pat, hint=hint)
        found = [c for c in self.columns if pat.match(c)]
        return found

    def search_columns(self, pat, hint='glob'):
        from kwutil import util_pattern
        pat = util_pattern.Pattern.coerce(pat, hint=hint)
        found = [c for c in self.columns if pat.search(c)]
        return found

    def subframe(self, key, drop_prefix=True):
        """
        Given a prefix key, return the subet columns that match it with the
        stripped prefix.
        """
        lookup_keys = []
        new_keys = []
        key_parts = key.split('.')
        nlevel = len(key_parts)
        for c in self.columns:
            path = c.split('.')
            if path[0:nlevel] == key_parts:
                lookup_keys.append(c)
                if drop_prefix:
                    new_keys.append('.'.join(path[nlevel:]))
        new = self.loc[:, lookup_keys]
        if drop_prefix:
            new.rename(ub.dzip(lookup_keys, new_keys), inplace=True, axis=1)
        return new

    def __getitem__(self, cols):
        if isinstance(cols, str):
            if cols not in self.columns:
                cols = self.query_column(cols)
                if not cols:
                    print(f'Available columns={self.columns}')
                    raise KeyError
        elif isinstance(cols, list):
            cols = list(ub.flatten([self.query_column(c) for c in cols]))
        return super().__getitem__(cols)


def pandas_add_prefix(data, prefix):
    return data.add_prefix(prefix)
    # mapper = {c: prefix + c for c in data.columns}
    # return data.rename(mapper, axis=1)


def aggregate_columns(df, aggregator=None, fallback='const',
                      nonconst_policy='error'):
    """
    Aggregates parameter columns based on per-column strategies / functions
    specified in ``aggregator``.

    Args:
        hash_cols (None | List[str]):
            columns whos values should be hashed together.

        aggregator (Dict[str, str | callable]):
            a dictionary mapping column names to a callable function that
            should be used to aggregate them. There a special string codes that
            we accept as well.
            Special functions are: hist, hash, min-max, const,

        fallback (str | callable):
            Aggregator function for any column without an explicit aggregator.
            Defaults to "const", which passes one value from the columns
            through if they are constant. If they are not constant, the
            nonconst-policy is triggered.

        nonconst_policy (str):
            Behavior when the aggregator is "const", but the input is
            non-constant. The policies are:
                * 'error' - error if unhandled non-uniform columns exist
                * 'drop' - remove unhandled non-uniform columns

    Returns:
        pd.Series

    TODO:
        - [ ] optimize this

    CommandLine:
        xdoctest -m geowatch.utils.util_pandas aggregate_columns

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> import numpy as np
        >>> num_rows = 10
        >>> columns = {
        >>>     'nums1': np.random.rand(num_rows),
        >>>     'nums2': np.random.rand(num_rows),
        >>>     'nums3': (np.random.rand(num_rows) * 10).astype(int),
        >>>     'nums4': (np.random.rand(num_rows) * 10).astype(int),
        >>>     'cats1': np.random.randint(0, 3, num_rows),
        >>>     'cats2': np.random.randint(0, 3, num_rows),
        >>>     'cats3': np.random.randint(0, 3, num_rows),
        >>>     'const1': ['a'] * num_rows,
        >>>     'strs1': [np.random.choice(list('abc')) for _ in range(num_rows)],
        >>> }
        >>> df = pd.DataFrame(columns)
        >>> aggregator = ub.udict({
        >>>     'nums1': 'mean',
        >>>     'nums2': 'max',
        >>>     'nums3': 'min-max',
        >>>     'nums4': 'stats',
        >>>     'cats1': 'histogram',
        >>>     'cats3': 'first',
        >>>     'cats2': 'hash12',
        >>>     'strs1': 'hash12',
        >>> })
        >>> #
        >>> # Test that the const fallback works
        >>> row = aggregate_columns(df, aggregator, fallback='const')
        >>> print('row = {}'.format(ub.urepr(row.to_dict(), nl=1)))
        >>> assert row['const1'] == 'a'
        >>> row = aggregate_columns(df.iloc[0:1], aggregator, fallback='const')
        >>> assert row['const1'] == 'a'
        >>> #
        >>> # Test that the drop fallback workds
        >>> row = aggregate_columns(df, aggregator, fallback='drop')
        >>> print('row = {}'.format(ub.urepr(row.to_dict(), nl=1)))
        >>> assert 'const1' not in row
        >>> row = aggregate_columns(df.iloc[0:1], aggregator, fallback='drop')
        >>> assert 'const1' not in row
        >>> #
        >>> # Test that non-constant policy triggers
        >>> aggregator_ = aggregator - {'cats3'}
        >>> import pytest
        >>> with pytest.raises(NonConstantError):
        >>>     row = aggregate_columns(df, aggregator_, nonconst_policy='error')
        >>> row = aggregate_columns(df, aggregator_, nonconst_policy='drop')
        >>> assert 'cats3' not in row
        >>> row = aggregate_columns(df, aggregator_, nonconst_policy='hash')
        >>> assert 'cats3' in row
        >>> #
        >>> # Test an empty dataframe returns an empty series
        >>> row = aggregate_columns(df.iloc[0:0], aggregator)
        >>> assert len(row) == 0
        >>> #
        >>> # Test single column cases work fine.
        >>> for col in df.columns:
        ...     subdf = df[[col]]
        ...     subagg = aggregate_columns(subdf, aggregator, fallback='const')
        ...     assert len(subagg) == 1
        >>> #
        >>> # Test single column drop case works
        >>> subagg = aggregate_columns(df[['cats3']], aggregator_, fallback='const', nonconst_policy='drop')
        >>> assert len(subagg) == 0
        >>> subagg = aggregate_columns(df[['cats3']], aggregator_, fallback='drop')
        >>> assert len(subagg) == 0

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> import numpy as np
        >>> num_rows = 10
        >>> columns = {
        >>>     'dates': ['2101-01-01', '1970-01-01', '2000-01-01'],
        >>>     'lists': [['a'], ['a', 'b'], []],
        >>>     'nums':  [1, 2, 3],
        >>> }
        >>> df = pd.DataFrame(columns)
        >>> aggregator = ub.udict({
        >>>     'dates': 'min-max',
        >>>     'lists': 'hash',
        >>>     'nums':  'mean',
        >>> })
        >>> row = aggregate_columns(df, aggregator)
        >>> print('row = {}'.format(ub.urepr(row.to_dict(), nl=1)))

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> import numpy as np
        >>> num_rows = 10
        >>> columns = {
        >>>     'items': [['a'], ['bcd', 'ef'], [], ['3', '234', '2343']],
        >>> }
        >>> df = pd.DataFrame(columns)
        >>> row = aggregate_columns(df, 'last', fallback='const')
        >>> columns = {
        >>>     'items': ['a', 'c', 'c', 'd'],
        >>>     'items2': [['a'], ['bcd', 'ef'], [], ['3', '234', '2343']],
        >>> }
        >>> df = pd.DataFrame(columns)
        >>> row = aggregate_columns(df, 'unique')
    """
    import pandas as pd
    # import numpy as np
    if len(df) == 0:
        return pd.Series(dtype=object)

    if aggregator is None:
        aggregator = {}
    if isinstance(aggregator, str):
        # If given as a string apply the aggreagatr to all columns
        aggregator = {c: aggregator for c in df.columns}

    aggregator = ub.udict(aggregator)

    # Handle columns that can be aggregated
    aggregated = []
    handled_keys = df.columns.intersection(aggregator.keys())
    unhandled_keys = df.columns.difference(handled_keys)

    if len(df) == 1 and fallback == 'const':
        agg_row = df.iloc[0]
        return agg_row
    elif len(df) == 1 and fallback == 'drop':
        agg_row = df.iloc[0][handled_keys]
        return agg_row
    else:
        aggregator = aggregator & handled_keys
        op_to_cols = ub.group_items(aggregator.keys(), aggregator.values())
        if len(unhandled_keys):
            op_to_cols[fallback] = unhandled_keys

        # Handle all columns with the same aggregator in a single call.
        for agg_op, cols in op_to_cols.items():
            toagg = df[cols]
            # toagg = toagg.select_dtypes(include=np.number)
            if isinstance(agg_op, str):
                agg_op_norm = SpecialAggregators.normalize_special_key(agg_op)
                if agg_op_norm == 'drop':
                    continue
                elif agg_op_norm == 'const':
                    # Special case where we will skip aggregation
                    part = _handle_const(toagg, nonconst_policy)
                    aggregated.append(part)
                    continue
                else:
                    agg_op = SpecialAggregators.special_lut.get(agg_op_norm, agg_op)

            # Using apply instead of pandas aggregate because we are allowed to
            # return a list result and have that be a single cell.
            part = toagg.apply(agg_op, result_type='reduce')
            # old: part = toagg.aggregate(agg_op)

            aggregated.append(part)
        if len(aggregated):
            agg_parts = pd.concat(aggregated)
            agg_row = agg_parts
        else:
            agg_row = pd.Series(dtype=object)
    return agg_row


def _handle_const(toagg, nonconst_policy):
    # Check which of the columns are actually constant
    is_const_cols = {
        k: ub.allsame(vs, eq=nan_eq)
        for k, vs in toagg.T.iterrows()}
    nonconst_cols = [k for k, v in is_const_cols.items() if not v]
    if nonconst_cols:
        if nonconst_policy == 'drop':
            toagg = toagg.iloc[0:1].drop(nonconst_cols, axis=1)
        elif nonconst_policy == 'error':
            raise NonConstantError(f'Values are non-constant in columns: {nonconst_cols}')
        elif nonconst_policy == 'hash':
            nonconst_data = toagg[nonconst_cols]
            const_part = toagg.iloc[0:1].drop(nonconst_cols, axis=1).iloc[0]
            nonconst_part = nonconst_data.apply(SpecialAggregators.hash, result_type='reduce')
            part = pd.concat([const_part, nonconst_part])
            return part
        else:
            raise KeyError(nonconst_policy)
    part = toagg.iloc[0]
    return part


class SpecialAggregators:

    def hash(x):
        return ub.hash_data(x.values.tolist())

    def hash12(x):
        return ub.hash_data(x.values.tolist())[0:12]

    def unique(x):
        try:
            return list(ub.unique(x))
        except Exception:
            return list(ub.unique(x, key=ub.hash_data))

    def min_max(x):
        return (x.min(), x.max())
        # ret = {
        #     'min': x.min(),
        #     'max': x.max(),
        # }
        # return ret

    @staticmethod
    def normalize_special_key(k):
        return k.replace('-', '_')

    special_lut = {
        'hash': hash,
        'hash12': hash12,
        'min_max': min_max,
        'stats': kwarray.stats_dict,
        'hist': ub.dict_hist,
        'unique': unique,
        'histogram': ub.dict_hist,
        'first': lambda x: x.iloc[0],
        'last': lambda x: x.iloc[-1],
    }


class NonConstantError(ValueError):
    ...


def nan_eq(a, b):
    if isinstance(a, float) and isinstance(b, float) and math.isnan(a) and math.isnan(b):
        return True
    else:
        return a == b


# Fix pandas groupby so it uses the new behavior with a list of len 1

class GroupbyFutureWrapper(wrapt.ObjectProxy):
    """
    Wraps a groupby object to get the new behavior sooner.

    TODO:
        - [ ] remove this when pandas 1.x no longer supported
    """

    def __iter__(self):
        keys = self.keys
        if isinstance(keys, list) and len(keys) == 1:
            # Handle this special case to avoid a warning
            for key, group in self.grouper.get_iterator(self._selected_obj, axis=self.axis):
                yield (key,), group
        else:
            # Otherwise use the parent impl
            yield from self.__wrapped__.__iter__()


def _fix_groupby(groups):
    """
    TODO:
        - [ ] remove this when pandas 1.x no longer supported
    """
    keys = groups.keys
    if isinstance(keys, list) and len(keys) == 1:
        return GroupbyFutureWrapper(groups)
    else:
        return groups


def pandas_fixed_groupby(df, by=None, **kwargs):
    """
    Fixed groupby behavior so length-one arguments are handled correctly

    Args:
        df (DataFrame):
        ** kwargs: groupby kwargs

    Example:
        >>> from geowatch.utils.util_pandas import *  # NOQA
        >>> df = pd.DataFrame({
        >>>     'Animal': ['Falcon', 'Falcon', 'Parrot', 'Parrot'],
        >>>     'Color': ['Blue', 'Blue', 'Blue', 'Yellow'],
        >>>     'Max Speed': [380., 370., 24., 26.]
        >>>     })
        >>> # Old behavior: In pandas 1.x groupbing by a legth-one list
        >>> # would return a single item instead of a length-one tuple
        >>> # as a key. In pandas 2.x this changed.
        >>> old1 = dict(list(df.groupby(['Animal', 'Color'])))
        >>> # In 1.x the keys will be a str, In 2.x the keys will be a List[str] for old2.
        >>> old2 = dict(list(df.groupby(['Animal'])))
        >>> old3 = dict(list(df.groupby('Animal')))
        >>> new1 = dict(list(pandas_fixed_groupby(df, ['Animal', 'Color'])))
        >>> # In 1.x and 2.x the keys will alwyas be a List[str] for new2.
        >>> new2 = dict(list(pandas_fixed_groupby(df, ['Animal'])))
        >>> new3 = dict(list(pandas_fixed_groupby(df, 'Animal')))
        >>> assert sorted(new1.keys())[0] == ('Falcon', 'Blue')
        >>> assert sorted(old1.keys())[0] == ('Falcon', 'Blue')
        >>> assert sorted(new3.keys())[0] == 'Falcon'
        >>> assert sorted(old3.keys())[0] == 'Falcon'
        >>> # This is the case that is fixed.
        >>> assert sorted(new2.keys())[0] == ('Falcon',)
        >>> import numpy as np
        >>> if np.lib.NumpyVersion(pd.__version__) < '2.0.0':
        >>>     assert sorted(old2.keys())[0] == 'Falcon'
    """
    groups = df.groupby(by=by, **kwargs)
    if PANDAS_GE_2_0_0:
        # No need to fix 2.x
        return groups
    else:
        fixed_groups = _fix_groupby(groups)
        return fixed_groups



import torch


def iou(y_pred, y_true, threshold):
    assert len(y_pred.shape) == len(y_true.shape) == 2, "Input tensor shapes should be (N, .)"
    mask_pred = threshold < y_pred
    mask_true = threshold < y_true
    intersection = torch.sum(mask_pred * mask_true, dim=-1)
    union = torch.sum(mask_pred + mask_true, dim=-1)
    r = intersection.float() / union.float()
    r[union == 0] = 1
    return r


def dice_loss(y_pred, y_true, smooth=1, eps=1e-7):
    """

    @param y_pred: (N, C, H, W)
    @param y_true: (N, C, H, W)
    @param smooth:
    @param eps:
    @return: (N, C)
    """
    numerator = 2 * torch.sum(y_true * y_pred, dim=(-1, -2))
    denominator = torch.sum(y_true, dim=(-1, -2)) + torch.sum(y_pred, dim=(-1, -2))
    return 1 - (numerator + smooth) / (denominator + smooth + eps)


def main():
    # import kornia
    # spatial_gradient_function = kornia.filters.SpatialGradient()
    #
    # image = torch.zeros((7, 7))
    # image[2:5, 2:5] = 1
    # print(image)
    #
    # grads = spatial_gradient_function(image[None, None, ...])[0, 0, ...] / 4
    # print(grads[0])
    # print(grads[1])

    y_true = torch.tensor([
        [0, 0, 0, 0, 1, 1, 1],
        [0, 0, 0, 0, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0]
    ])
    y_pred = torch.tensor([
        [0, 0, 0, 0, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0],
    ])
    print(y_true.shape)
    print(y_pred.shape)
    r = iou(y_pred, y_true, threshold=0.5)
    print(r)


if __name__ == "__main__":
    main()



# FIXME:
# Adds the "modules" subdirectory to the python path.
# See https://gitlab.kitware.com/smart/watch/-/merge_requests/148#note_1050127
# for discussion of how to refactor this in the future.
import geowatch_tpl  # NOQA

import warnings
import numpy as np
import torch
import torch.nn as nn
import torchvision.transforms

import pytorch_lightning as pl

from .backbone import get_backbone

from frame_field_learning import data_transforms
from frame_field_learning import local_utils
from frame_field_learning.model_multi import Multi_FrameFieldModel

from scipy import ndimage

dfactor = 25.5

# device = torch.device("cuda" if torch.cuda.is_available() else "cpu") # PyTorch v0.4.0


#-------------------------------------------
# Modify the batch_norm layers
#-------------------------------------------

def modify_bn(model, track_running_stats=True, bn_momentum=0.1):
    for m in model.modules():
        for child in m.children():
            if isinstance(child, nn.BatchNorm2d):

                child.momentum = bn_momentum
                child.track_running_stats = track_running_stats

                if track_running_stats is False:
                    child.running_mean = None
                    child.running_var =  None

    return model


#-------------------------------------------------
# Depth/Label/Shadow/Facade Eestimation Module
#-------------------------------------------------

class MultiTaskModel(pl.LightningModule):

    def __init__(
        self,
        batch_size: int = 1,
        checkpoint: str = None,
        config: dict = None,
        test_img_dir: str = None,
        test_img_list: str = None,
        gpus: str = '0',
        **kwargs,
    ):

        super().__init__(**kwargs)
        self.gpus = gpus
        self.checkpoint = checkpoint
        self.batch_size = batch_size

        self.test_img_dir = test_img_dir
        self.test_img_list = test_img_list

        self.config = config

        self.backbone = get_backbone(self.config["backbone_params"])

        train_online_cuda_transform = None
        eval_online_cuda_transform = None

        self.net = Multi_FrameFieldModel(
            self.config,
            backbone=self.backbone,
            train_transform=train_online_cuda_transform,
            eval_transform=eval_online_cuda_transform)

        self.transform =  data_transforms.get_online_cuda_transform(
            self.config,
            augmentations=self.config["data_aug_params"]["enable"])

    def forward(self, x, tta=False):
        return self.net(x, tta)

    def test_step(self, batch, batch_idx):

        out_arr = []
        for i, image in enumerate(batch):
            if isinstance(image, dict):
                gid = image['id']
                # img_info = image
                image = image['imgdata']

            with torch.no_grad():

                image_float = image / 255.0
                mean = np.mean(image_float.reshape(-1, image_float.shape[-1]), axis=0)
                std = np.std(image_float.reshape(-1, image_float.shape[-1]), axis=0)

                batch2 = {
                    "image": torchvision.transforms.functional.to_tensor(image)[None, ...],
                    "image_mean": torch.from_numpy(mean)[None, ...],
                    "image_std": torch.from_numpy(std)[None, ...],
                }

                batch2 = local_utils.batch_to_cuda(batch2)

                pred2, batch2 = self(batch2, tta=True)

                output_depth = pred2['depth'][0, 0, :, :].cpu().data.numpy()
                output_label = pred2['seg'][0, 0, :, :].cpu().data.numpy()

                weighted_depth = dfactor * output_depth

                alpha = 0.9
                weighted_seg = alpha * output_label + (1.0 - alpha) * np.minimum(0.99, weighted_depth / 70.0)

                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    tmp2 = 255 * anisotropic_diffusion(weighted_seg, niter=1, kappa=100, gamma=0.8)
                weighted_final = ndimage.median_filter(tmp2.astype(np.uint8), size=7)

                # Image.fromarray(weighted_final.astype(np.uint8)).save('/output/weighted_final.png')

            out_arr.append((gid, weighted_final))
        return out_arr

    @staticmethod
    def add_model_specific_args(parent_parser):  # pragma: no-cover
        parser = parent_parser.add_argument_group("MultiTaskModel")

        parser.add_argument('--checkpoint', default=None, type=str,
                            help='checkpoint to use for testing')
        parser.add_argument('--config', '--config', default=None, type=str,
                            help='Name of the config file, excluding the .json file extension.')
        parser.add_argument('--test_img_dir', '--test_img_dir', default=None, type=str,
                            help='directory where test images are located')
        parser.add_argument('--test_img_list', '--test_img_list', default=None, type=str,
                            help='list of test images')
        parser.add_argument('--gpus', default='0', type=str,
                            help='GPU')

        return parent_parser


# Vendored in to deal with 3.10 issue
def anisotropic_diffusion(img, niter=1, kappa=50,
                          gamma=0.1, voxelspacing=None, option=1):
    r"""
    Edge-preserving, XD Anisotropic diffusion.


    Parameters
    ----------
    img : array_like
        Input image (will be cast to numpy.float).
    niter : integer
        Number of iterations.
    kappa : integer
        Conduction coefficient, e.g. 20-100. ``kappa`` controls conduction
        as a function of the gradient. If ``kappa`` is low small intensity
        gradients are able to block conduction and hence diffusion across
        steep edges. A large value reduces the influence of intensity gradients
        on conduction.
    gamma : float
        Controls the speed of diffusion. Pick a value :math:`<= .25` for stability.
    voxelspacing : tuple of floats or array_like
        The distance between adjacent pixels in all img.ndim directions
    option : {1, 2, 3}
        Whether to use the Perona Malik diffusion equation No. 1 or No. 2,
        or Tukey's biweight function.
        Equation 1 favours high contrast edges over low contrast ones, while
        equation 2 favours wide regions over smaller ones. See [1]_ for details.
        Equation 3 preserves sharper boundaries than previous formulations and
        improves the automatic stopping of the diffusion. See [2]_ for details.

    Returns
    -------
    anisotropic_diffusion : ndarray
        Diffused image.

    Notes
    -----
    Original MATLAB code by Peter Kovesi,
    School of Computer Science & Software Engineering,
    The University of Western Australia,
    pk @ csse uwa edu au,
    <http://www.csse.uwa.edu.au>

    Translated to Python and optimised by Alistair Muldal,
    Department of Pharmacology,
    University of Oxford,
    <alistair.muldal@pharm.ox.ac.uk>

    Adapted to arbitrary dimensionality and added to the MedPy library Oskar Maier,
    Institute for Medical Informatics,
    Universitaet Luebeck,
    <oskar.maier@googlemail.com>

    June 2000  original version. -
    March 2002 corrected diffusion eqn No 2. -
    July 2012 translated to Python -
    August 2013 incorporated into MedPy, arbitrary dimensionality -

    References
    ----------
    .. [1] P. Perona and J. Malik.
       Scale-space and edge detection using ansotropic diffusion.
       IEEE Transactions on Pattern Analysis and Machine Intelligence,
       12(7):629-639, July 1990.
    .. [2] M.J. Black, G. Sapiro, D. Marimont, D. Heeger
       Robust anisotropic diffusion.
       IEEE Transactions on Image Processing,
       7(3):421-432, March 1998.
    """
    # define conduction gradients functions
    import numpy
    if option == 1:
        def condgradient(delta, spacing):
            return numpy.exp(-(delta / kappa)**2.) / float(spacing)
    elif option == 2:
        def condgradient(delta, spacing):
            return 1. / (1. + (delta / kappa)**2.) / float(spacing)
    elif option == 3:
        kappa_s = kappa * (2**0.5)

        def condgradient(delta, spacing):
            top = 0.5 * ((1. - (delta / kappa_s)**2.)**2.) / float(spacing)
            return numpy.where(numpy.abs(delta) <= kappa_s, top, 0)

    # initialize output array
    out = numpy.array(img, dtype=numpy.float32, copy=True)

    # set default voxel spacing if not supplied
    if voxelspacing is None:
        voxelspacing = tuple([1.] * img.ndim)

    # initialize some internal variables
    deltas = [numpy.zeros_like(out) for _ in range(out.ndim)]

    for _ in range(niter):

        # calculate the diffs
        for i in range(out.ndim):
            slicer = tuple([slice(None, -1) if j == i else slice(None)
                           for j in range(out.ndim)])
            deltas[i][slicer] = numpy.diff(out, axis=i)

        # update matrices
        matrices = [
            condgradient(
                delta,
                spacing) *
            delta for delta,
            spacing in zip(
                deltas,
                voxelspacing)]

        # subtract a copy that has been shifted ('Up/North/West' in 3D case) by one
        # pixel. Don't as questions. just do it. trust me.
        for i in range(out.ndim):
            slicer = tuple([slice(1, None) if j == i else slice(None)
                           for j in range(out.ndim)])
            matrices[i][slicer] = numpy.diff(matrices[i], axis=i)

        # update the image
        out += gamma * (numpy.sum(matrices, axis=0))

    return out



from tqdm import tqdm

import torch

from lydorn_utils import image_utils
from lydorn_utils import polygon_utils


class ToPatches(object):
    """Splits sample into patches"""

    def __init__(self, stride, size):
        self.stride = stride
        self.size = size

    def _to_patch(self, sample):
        image = sample["image"]
        gt_polygons = sample["gt_polygons"]

        patch_boundingboxes = image_utils.compute_patch_boundingboxes(image.shape[0:2],
                                                                      stride=self.stride,
                                                                      patch_res=self.size)

        patches = []
        for patch_boundingbox in tqdm(patch_boundingboxes, desc="Patching", leave=False):
            # Crop image
            patch_image = image[patch_boundingbox[0]:patch_boundingbox[2], patch_boundingbox[1]:patch_boundingbox[3], :]
            patch_gt_polygon = polygon_utils.crop_polygons_to_patch_if_touch(gt_polygons, patch_boundingbox)
            if len(patch_gt_polygon) == 0:
                patch_gt_polygon = None
            sample["image"] = patch_image
            sample["gt_polygons"] = patch_gt_polygon
            sample["patch_bbox"] = torch.tensor(patch_boundingbox)
            patches.append(sample.copy())
        return patches

    def __call__(self, data_list):
        patch_list = []
        for data in data_list:
            patches = self._to_patch(data)
            patch_list.extend(patches)
        return patch_list



def test_dynamic_resolution():
    from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    import geowatch
    import ubelt as ub

    # Ensure the two videos have different sizes
    # (one is much larger than the other)
    coco_dset1 = geowatch.coerce_kwcoco(
        'geowatch', num_videos=1, image_size=(8, 8), num_frames=3,
        multisensor=False,
        multispectral=False,
        geodata={
            'enabled': True,
            'region_geom': 'random-proportional',
            'target_gsd': 2.0
        }, rng=10)
    coco_dset2 = geowatch.coerce_kwcoco(
        'geowatch', num_videos=1, image_size=(64, 64), num_frames=3,
        multisensor=False,
        multispectral=False,
        geodata={
            'enabled': True,
            'region_geom': 'random-proportional',
            'target_gsd': 2.0
        }, rng=11)
    coco_dset = coco_dset1.union(coco_dset2)

    coco_dset.images().lookup(['width', 'height'])
    coco_dset.videos().lookup(['width', 'height'])

    # Enable "dynamic fixed resolution" which should scale down the image
    # so there are a maximum number of windows
    dynamic_fixed_resolution = {
        'max_winspace_full_dims': (4, 4),
    }
    # dynamic_fixed_resolution = None
    self = KWCocoVideoDataset(coco_dset, time_dims=3,
                              window_dims=(4, 4), fixed_resolution='4.0GSD',
                              channels='red|green|blue', autobuild=False, mode='test',
                              use_grid_cache=False,
                              resample_invalid_frames=False,
                              force_bad_frames=True,
                              dynamic_fixed_resolution=dynamic_fixed_resolution)
    self.requested_tasks['change'] = False
    self.requested_tasks['class'] = False
    self.requested_tasks['boxes'] = False
    self.requested_tasks['outputs'] = False
    self._init()
    target1 = (self.sample_grid['targets'][0])
    target2 = (self.sample_grid['targets'][-1])

    # Check that we downsample the big video but keep the small video to some
    # target resolution at its regular resolution.
    import rich
    rich.print(f'target1 = {ub.urepr(target1, nl=1)}')
    rich.print(f'target2 = {ub.urepr(target2, nl=1)}')

    sample1 = self[target1]
    sample2 = self[target2]
    summary1 = self.summarize_item(sample1)
    summary2 = self.summarize_item(sample2)
    import rich
    rich.print(f'summary1 = {ub.urepr(summary1, nl=-2)}')
    rich.print(f'summary2 = {ub.urepr(summary2, nl=-2)}')

    # The dynamic resolution should force the samples to be the same size
    # in the sampled input dimensions, but the first should be at a coarser
    # resolution.
    shape1 = sample1['frames'][0]['modes']['red|green|blue'].shape
    shape2 = sample2['frames'][0]['modes']['red|green|blue'].shape
    assert tuple(shape1) == tuple(shape2)

    # Disable test for the disussed reason. FIXME.
    if False:
        # Note: the random number generator which controls some details of the
        # generated video / geo-crs size (even though we specify the image size)
        # influences if the following statement is true.  If this starts to fail,
        # it could be due to RNG seed issues.  It would be nice to ensure this test
        # properly generates all relevant aspects of the data we are trying to
        # test, and that those are documented well.
        scale1 = sample1['frames'][0]['scale_outspace_from_vid']
        scale2 = sample2['frames'][0]['scale_outspace_from_vid']
        import numpy as np
        assert np.all(scale1 < scale2)


def distance_weights():
    from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    import ubelt as ub
    # Demo toy data without augmentation
    import kwcoco
    import numpy as np
    #src = geowatch.coerce_kwcoco('geowatch-msi', geodata=True, dates=True)
    coco_dset = kwcoco.CocoDataset.demo('vidshapes2-multispectral', num_frames=10)
    channels = 'B10,B8a|B1,B8'
    self = KWCocoVideoDataset(coco_dset, time_dims=4, window_dims=(300, 300),
                              channels=channels,
                              input_space_scale='native',
                              output_space_scale=None,
                              window_space_scale=1.2,
                              augment_space_shift_rate=0.5,
                              use_grid_negatives=False,
                              use_grid_positives=False,
                              use_centered_positives=True,
                              absolute_weighting=True,
                              time_sampling='uniform',
                              time_kernel='-1year,0,1month,1year',
                              modality_dropout=0.5,
                              channel_dropout=0.5,
                              temporal_dropout=0.7,
                              temporal_dropout_rate=1.0)
    # Add weights to annots
    annots = self.sampler.dset.annots()
    annots.set('weight', 2 + np.random.rand(len(annots)) * 10)
    self.disable_augmenter = False
    # Summarize batch item in text
    index = self.sample_grid['targets'][self.sample_grid['positives_indexes'][3]]
    item = self[index]
    summary = self.summarize_item(item)
    print('item summary: ' + ub.urepr(summary, nl=2))
    # Draw batch item
    canvas = self.draw_item(item, draw_weights=True)
    # xdoctest: +REQUIRES(--show)
    import kwplot
    kwplot.autompl()
    kwplot.imshow(canvas)
    kwplot.show_if_requested()


def test_msi_auto_channels():
    from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    import kwcoco
    import ubelt as ub

    coco_msi = kwcoco.CocoDataset.demo('vidshapes2-msi', num_frames=1)
    coco_rgb = kwcoco.CocoDataset.demo('vidshapes2', num_frames=1)

    msi_spec = coco_msi.coco_image(1).channels.spec
    assert 'B1' in msi_spec, 'kwcoco should provide this'
    assert 'B8' in msi_spec, 'kwcoco should provide this'
    assert 'B8a' in msi_spec, 'kwcoco should provide this'

    sample_kwargs = {
        'time_dims': 1,
        'window_dims': (128, 128)
    }

    datasets = {}
    # Check that regular RGB works fine
    datasets['rgb_auto'] = KWCocoVideoDataset(coco_rgb, **sample_kwargs)
    datasets['rgb_explicit'] = KWCocoVideoDataset(coco_rgb, **sample_kwargs,
                                                  channels='r|g|b')
    datasets['msi_explicit'] = KWCocoVideoDataset(coco_msi, **sample_kwargs,
                                                  channels='B10,B8a|B1,B8')
    datasets['msi_auto'] = KWCocoVideoDataset(coco_msi, **sample_kwargs)

    if 0:
        # First check that the matching sensorchans does what we expect
        # This was broken in delayed image 0.4.2 and fixed in 0.4.3
        datasets['msi_explicit'].sample_sensorchan.matching_sensor('sensor1')

    results = []
    errors = []
    for key, dataset in datasets.items():
        row = {'key': key}
        try:
            item = dataset.getitem(0)
        except Exception as ex:
            row['status'] = 'fail'
            row['ex'] = str(ex)[0:16]
            errors.append(row)
            raise
        else:
            row['item'] = str(item)[0:32]
            row['status'] = 'pass'
        results.append(row)

    if errors:
        import pandas as pd
        import rich
        df = pd.DataFrame(results)
        rich.print(df.to_string())
        for key, dataset in datasets.items():
            print('----')
            print(f'key={key}')
            print(f'dataset.sensorchan        = {ub.urepr(dataset.sensorchan, nl=1)}')
            print(f'dataset.sample_sensorchan = {ub.urepr(dataset.sample_sensorchan, nl=1)}')
            print(f'dataset.input_sensorchan  = {ub.urepr(dataset.input_sensorchan, nl=1)}')
        raise AssertionError('Problems in dataset getitems, maybe something to do with sensorchan configs?')


if __name__ == "__main__":
    distance_weights()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deeplab2/dataset.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='deeplab2/dataset.proto',
    package='deeplab2',
    syntax='proto2',
    serialized_options=None,
    serialized_pb=b'\n\x16\x64\x65\x65plab2/dataset.proto\x12\x08\x64\x65\x65plab2\"\xc2\x05\n\x0e\x44\x61tasetOptions\x12\x0f\n\x07\x64\x61taset\x18\x01 \x01(\t\x12\x14\n\x0c\x66ile_pattern\x18\x02 \x03(\t\x12\x16\n\nbatch_size\x18\x03 \x01(\x05:\x02\x33\x32\x12\x11\n\tcrop_size\x18\x04 \x03(\x05\x12\x18\n\x10min_resize_value\x18\x05 \x03(\x05\x12\x18\n\x10max_resize_value\x18\x06 \x03(\x05\x12\x15\n\rresize_factor\x18\x07 \x01(\x05\x12\x43\n\raugmentations\x18\x08 \x01(\x0b\x32,.deeplab2.DatasetOptions.AugmentationOptions\x12\x10\n\x05sigma\x18\t \x01(\x02:\x01\x38\x12.\n\x1fincrease_small_instance_weights\x18\n \x01(\x08:\x05\x66\x61lse\x12&\n\x18small_instance_threshold\x18\x0b \x01(\x05:\x04\x34\x30\x39\x36\x12 \n\x15small_instance_weight\x18\x0c \x01(\x02:\x01\x33\x12\x1d\n\x0euse_two_frames\x18\r \x01(\x08:\x05\x66\x61lse\x12&\n\x18\x64\x65\x63ode_groundtruth_label\x18\x0e \x01(\x08:\x04true\x12(\n\x19thing_id_mask_annotations\x18\x0f \x01(\x08:\x05\x66\x61lse\x12\x19\n\x0cmax_thing_id\x18\x10 \x01(\x05:\x03\x31\x32\x38\x12\x1d\n\x0euse_next_frame\x18\x11 \x01(\x08:\x05\x66\x61lse\x1a\x96\x01\n\x13\x41ugmentationOptions\x12\x1b\n\x10min_scale_factor\x18\x01 \x01(\x02:\x01\x31\x12\x1b\n\x10max_scale_factor\x18\x02 \x01(\x02:\x01\x31\x12$\n\x16scale_factor_step_size\x18\x03 \x01(\x02:\x04\x30.25\x12\x1f\n\x17\x61utoaugment_policy_name\x18\x04 \x01(\t'
)

_DATASETOPTIONS_AUGMENTATIONOPTIONS = _descriptor.Descriptor(
    name='AugmentationOptions',
    full_name='deeplab2.DatasetOptions.AugmentationOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='min_scale_factor', full_name='deeplab2.DatasetOptions.AugmentationOptions.min_scale_factor', index=0,
            number=1, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_scale_factor', full_name='deeplab2.DatasetOptions.AugmentationOptions.max_scale_factor', index=1,
            number=2, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='scale_factor_step_size',
            full_name='deeplab2.DatasetOptions.AugmentationOptions.scale_factor_step_size', index=2,
            number=3, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.25),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='autoaugment_policy_name',
            full_name='deeplab2.DatasetOptions.AugmentationOptions.autoaugment_policy_name', index=3,
            number=4, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=593,
    serialized_end=743,
)

_DATASETOPTIONS = _descriptor.Descriptor(
    name='DatasetOptions',
    full_name='deeplab2.DatasetOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='dataset', full_name='deeplab2.DatasetOptions.dataset', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='file_pattern', full_name='deeplab2.DatasetOptions.file_pattern', index=1,
            number=2, type=9, cpp_type=9, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='batch_size', full_name='deeplab2.DatasetOptions.batch_size', index=2,
            number=3, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=32,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='crop_size', full_name='deeplab2.DatasetOptions.crop_size', index=3,
            number=4, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='min_resize_value', full_name='deeplab2.DatasetOptions.min_resize_value', index=4,
            number=5, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_resize_value', full_name='deeplab2.DatasetOptions.max_resize_value', index=5,
            number=6, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='resize_factor', full_name='deeplab2.DatasetOptions.resize_factor', index=6,
            number=7, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='augmentations', full_name='deeplab2.DatasetOptions.augmentations', index=7,
            number=8, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='sigma', full_name='deeplab2.DatasetOptions.sigma', index=8,
            number=9, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(8),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='increase_small_instance_weights', full_name='deeplab2.DatasetOptions.increase_small_instance_weights',
            index=9,
            number=10, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='small_instance_threshold', full_name='deeplab2.DatasetOptions.small_instance_threshold', index=10,
            number=11, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=4096,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='small_instance_weight', full_name='deeplab2.DatasetOptions.small_instance_weight', index=11,
            number=12, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(3),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_two_frames', full_name='deeplab2.DatasetOptions.use_two_frames', index=12,
            number=13, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='decode_groundtruth_label', full_name='deeplab2.DatasetOptions.decode_groundtruth_label', index=13,
            number=14, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='thing_id_mask_annotations', full_name='deeplab2.DatasetOptions.thing_id_mask_annotations', index=14,
            number=15, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_thing_id', full_name='deeplab2.DatasetOptions.max_thing_id', index=15,
            number=16, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=128,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_next_frame', full_name='deeplab2.DatasetOptions.use_next_frame', index=16,
            number=17, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[_DATASETOPTIONS_AUGMENTATIONOPTIONS, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=37,
    serialized_end=743,
)

_DATASETOPTIONS_AUGMENTATIONOPTIONS.containing_type = _DATASETOPTIONS
_DATASETOPTIONS.fields_by_name['augmentations'].message_type = _DATASETOPTIONS_AUGMENTATIONOPTIONS
DESCRIPTOR.message_types_by_name['DatasetOptions'] = _DATASETOPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DatasetOptions = _reflection.GeneratedProtocolMessageType('DatasetOptions', (_message.Message,), {

    'AugmentationOptions': _reflection.GeneratedProtocolMessageType('AugmentationOptions', (_message.Message,), {
        'DESCRIPTOR': _DATASETOPTIONS_AUGMENTATIONOPTIONS,
        '__module__': 'deeplab2.dataset_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.DatasetOptions.AugmentationOptions)
    })
    ,
    'DESCRIPTOR': _DATASETOPTIONS,
    '__module__': 'deeplab2.dataset_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.DatasetOptions)
})
_sym_db.RegisterMessage(DatasetOptions)
_sym_db.RegisterMessage(DatasetOptions.AugmentationOptions)

# @@protoc_insertion_point(module_scope)



# class Map(object):
#     """Applies parameter transform to all items of input list"""
#
#     def __init__(self, transform, multi_outs=False):
#         self.transform = transform
#         if multi_outs:
#             self.__call__ = self._map_multi_outs
#         else:
#             self.__call__ = self._map_single_out
#
#     def _map_single_out(self, data_list):
#         assert type(data_list) == list, "data_list should be a list"
#         return [self.transform(item) for item in data_list]
#
#     def _map_multi_outs(self, data_list):
#         assert type(data_list) == list, "data_list should be a list"
#         return tuple(zip(*[self.transform(item) for item in data_list]))


class Map(object):
    """Applies parameter transform to all items of input list"""

    def __init__(self, transform):
        self.transform = transform

    def __call__(self, data_list):
        assert isinstance(data_list, list), "data_list should be a list"
        return [self.transform(item) for item in data_list]



import pytest
import ubelt as ub
from geowatch.geoannots.geomodels import RegionModel
from geowatch.geoannots.geomodels import RegionHeader


def test_dump_load():
    from geowatch.geoannots import geomodels
    # Create arguments to the script
    dpath = ub.Path.appdir('geowatch/tests/geomodels/test_dump_load').delete().ensuredir()

    region = geomodels.RegionModel.random()
    fpath1 = dpath / 'region1.geojson'
    fpath2 = dpath / 'region2.geojson'

    region.dump(fpath1)

    with open(fpath2, 'w') as file:
        region.dump(file)

    assert fpath1.exists()
    assert fpath2.exists()

    recon1 = geomodels.RegionModel.load(fpath1)

    with open(fpath2, 'r') as file:
        recon2 = geomodels.RegionModel.load(file)

    assert recon1 == region
    assert recon2 == region


def test_dumps_loads():
    from geowatch.geoannots import geomodels
    region = geomodels.RegionModel.random()
    text = region.dumps()
    recon = region.loads(text)
    assert recon == region


def test_infer_region_header():
    # Case: RegionModel has None header info
    self = RegionModel.random()
    self.header['properties']['mgrs'] = None
    self.header['geometry'] = None
    assert self.header is not None
    self.infer_header()
    assert self.header is not None
    assert self.header['geometry'] is not None
    assert self.header['properties']['mgrs'] is not None

    # Case: RegionModel has missing header info
    self = RegionModel.random()
    del self.header['properties']['mgrs']
    del self.header['geometry']
    assert self.header is not None
    self.infer_header()
    assert self.header is not None
    assert self.header['geometry'] is not None
    assert self.header['properties']['mgrs'] is not None

    # Case: RegionModel has no header
    self = RegionModel.random()
    self.features.remove(self.header)
    assert self.header is None
    # Infer the header using site summaries
    self.infer_header()
    assert self.header is not None
    assert self.header['properties']['region_id'] is None, 'should not be able to infer region id here'

    # Case: RegionModel has no header, and a partial header is given
    self = RegionModel.random()
    self.features.remove(self.header)
    assert self.header is None
    # Infer the header using site summaries
    self.infer_header(RegionHeader(properties={'region_id': 'foobar'}))
    assert self.header is not None
    assert self.header['properties']['region_id'] == 'foobar'

    # Case: RegionModel has a header and a partial header is given
    self = RegionModel.random()
    with pytest.raises(ValueError):
        self.infer_header(RegionHeader(properties={'region_id': 'foobar'}))

    # Case: RegionModel has a header
    self = RegionModel.random()
    old_header = self.header
    old_header_str = str(old_header)
    self.infer_header()
    assert self.header is old_header, 'should not change anything'
    assert old_header_str == str(self.header), 'should not change anything'


def test_infer_region_header_conflict_mgrs():
    self1 = RegionModel.random(rng=123)
    self2 = RegionModel.random(rng=321)

    # Make a region model that spans multiple MGRS tiles
    self1.features.extend(list(self2.site_summaries()))

    # Remove the header
    self1.features.remove(self1.header)

    # Infered header should handle conflicting mgrs gracefully



# import sys, ubelt  # NOQA
# sys.path.append(ubelt.expandpath('~/code/loss-of-plasticity'))  # NOQA
import ubelt as ub
import rich
from collections import defaultdict
from lop.utils.AdamGnT import AdamGnT
from math import sqrt
from torch.nn import Conv2d, Linear
import numpy as np
import torch
from torch.nn.init import calculate_gain


def get_layer_bound(layer, init, gain):
    if isinstance(layer, Conv2d):
        return sqrt(1 / (layer.in_channels *
                    layer.kernel_size[0] * layer.kernel_size[1]))
    elif isinstance(layer, Linear):
        if init == 'default':
            bound = sqrt(1 / layer.in_features)
        elif init == 'xavier':
            bound = gain * sqrt(6 / (layer.in_features + layer.out_features))
        elif init == 'lecun':
            bound = sqrt(3 / layer.in_features)
        else:
            bound = gain * sqrt(3 / layer.in_features)
        return bound


KNOWN_LAYERS = (torch.nn.Linear, torch.nn.Conv2d)


def test_complex():
    """
    Ignore:
        import sys, ubelt
        sys.path.append(ubelt.expandpath('~/code/watch/geowatch_tpl/submodules/loss-of-plasticity'))
        sys.path.append(ubelt.expandpath('~/code/watch/geowatch_tpl/submodules/torchview'))
        from lop.algos.gen_and_test import *  # NOQA
    """
    from watch.tasks.fusion.methods.channelwise_transformer import MultimodalTransformer
    from watch.tasks.fusion import datamodules
    print('(STEP 0): SETUP THE DATA MODULE')
    datamodule = datamodules.KWCocoVideoDataModule(
        train_dataset='special:vidshapes-watch', num_workers=4, channels='auto')
    datamodule.setup('fit')
    dataset = datamodule.torch_datasets['train']
    print('(STEP 1): ESTIMATE DATASET STATS')
    dataset_stats = dataset.cached_dataset_stats(num=3)
    print('dataset_stats = {}'.format(ub.urepr(dataset_stats, nl=3)))
    loader = datamodule.train_dataloader()
    print('(STEP 2): SAMPLE BATCH')
    batch = next(iter(loader))
    for item_idx, item in enumerate(batch):
        print(f'item_idx={item_idx}')
        item_summary = dataset.summarize_item(item)
        print('item_summary = {}'.format(ub.urepr(item_summary, nl=2)))
    print('(STEP 3): THE REST OF THE TEST')
    #self = MultimodalTransformer(arch_name='smt_it_joint_p8')
    net = MultimodalTransformer(arch_name='smt_it_stm_p8',
                                dataset_stats=dataset_stats,
                                classes=datamodule.classes,
                                decoder='mlp', change_loss='dicefocal',
                                attention_impl='exact')

    meta = MetaNetwork(net)
    meta._build()
    input_data = batch[0:1]
    meta.trace(input_data=input_data)

    for nx_node, node_data in meta.nx_graph.nodes(data=True):
        if 'tv_compute_node' in node_data:
            if 'layer_name' in node_data:
                layer_name = node_data['layer_name']
                tv_node = node_data['tv_compute_node']
                tv_node._show_name = f'{layer_name}:{tv_node.name}'
                print(tv_node)
                print('tv_node._show_name = {}'.format(ub.urepr(tv_node._show_name, nl=1)))

    print('meta.tv_graph.edge_list = {}'.format(ub.urepr(meta.tv_graph.edge_list, nl=1)))
    meta.tv_graph.visual_graph.render(format='png')

    ### DEV
    found_connections = []
    missing_connections = []
    for layer_name in meta.layer_names:
        next_layer_name = list([t[0] for t in meta.next_layers(layer_name)])
        print(f'connections {layer_name} -> {next_layer_name}')
        if len(next_layer_name):
            found_connections.append(layer_name)
        else:
            missing_connections.append(layer_name)

    missing_connections

    for nx_node, node_data in meta.nx_graph.nodes(data=True):
        if 'layer_name' in node_data:
            layer_name = node_data['layer_name']
            if layer_name in found_connections:
                node_data['label'] = '[green]' + node_data['label'] + '[\\green]'
            elif layer_name in missing_connections:
                node_data['label'] = '[red]' + node_data['label'] + '[\\red]'
            else:
                ...

    import networkx as nx
    nx.write_network_text(meta.nx_graph, rich.print, vertical_chains=1, end='')

    for name in missing_connections:
        layer = meta.name_to_layer[name]
        print(f'{name} {layer}')

    from torch.optim import AdamW
    opt = AdamW(net.parameters())
    net(batch)
    self = GenerateAndTest(net, 'relu', opt, input_data)
    self.gen_and_test()

    # Do a forward pass so activations are populated
    net.train()
    for i in ub.ProgIter(range(100)):
        outputs = net(input_data)
        loss = outputs['loss']
        # loss = outputs.sum()
        loss.backward()
        opt.zero_grad()
        opt.step()
        self.gen_and_test()


class GenerateAndTest:
    """
    Generate-and-Test algorithm for feed forward neural networks, based on
    maturity-threshold based replacement

    CommandLine:
        xdoctest -m lop/algos/gen_and_test.py GenerateAndTest

    Example:
        >>> import geowatch_tpl
        >>> import torchvision
        >>> from lop.algos.gen_and_test import *  # NOQA
        >>> from torch.optim import AdamW
        >>> geowatch_tpl.import_submodule('torchview')
        >>> geowatch_tpl.import_submodule('lop')
        >>> net = torchvision.models.resnet18()
        >>> opt = AdamW(net.parameters())
        >>> hidden_activation = 'relu'
        >>> #inputs = torch.rand(2, 3, 224, 224)
        >>> #outputs = net(inputs)
        >>> #loss = outputs.sum()
        >>> #loss.backward()
        >>> #opt.step()
        >>> #opt.zero_grad()
        >>> input_data = torch.rand(2, 3, 224, 224)
        >>> self = GenerateAndTest(net, hidden_activation, opt, input_data=input_data)
        >>> meta = self.meta
        >>> import networkx as nx
        >>> nx.write_network_text(meta.nx_graph, vertical_chains=1)
        >>> for nx_node, node_data in meta.nx_graph.nodes(data=True):
        >>>     if 'tv_compute_node' in node_data:
        >>>         if 'layer_name' in node_data:
        >>>             layer_name = node_data['layer_name']
        >>>             tv_node = node_data['tv_compute_node']
        >>>             tv_node._show_name = f'{layer_name}:{tv_node.name}'
        >>>             print(tv_node)
        >>>             print('tv_node._show_name = {}'.format(ub.urepr(tv_node._show_name, nl=1)))
        >>> print('meta.tv_graph.edge_list = {}'.format(ub.urepr(meta.tv_graph.edge_list, nl=1)))
        >>> meta.tv_graph.fill_visual_graph()
        >>> meta.tv_graph.visual_graph.render(format='png')
        >>> # Do a forward pass so activations are populated
        >>> for i in ub.ProgIter(range(100)):
        >>>     outputs = net(input_data)
        >>>     loss = outputs.sum()
        >>>     loss.backward()
        >>>     #
        >>>     input_data2 = torch.rand(2, 3, 224, 224)
        >>>     outputs = net(input_data2)
        >>>     loss = outputs.sum()
        >>>     loss.backward()
        >>>     #
        >>>     input_data2 = torch.rand(2, 3, 224, 224)
        >>>     outputs = net(input_data2)
        >>>     loss = outputs.sum()
        >>>     loss.backward()
        >>>     #
        >>>     opt.zero_grad()
        >>>     opt.step()
        >>>     self.gen_and_test()

        opt.param_groups[0]['params']
    """

    def __init__(
            self,
            net,
            hidden_activation,
            opt,
            input_data,
            decay_rate=0.99,
            replacement_rate=1e-4,
            init='kaiming',
            device="cpu",
            maturity_threshold=20,
            util_type='contribution',
            num_last_filter_outputs=4,
            accumulate=False,
            verbose=1,
    ):
        super().__init__()
        self.device = device
        self.net = net
        self.verbose = verbose

        self.meta = MetaNetwork(net)
        if self.verbose:
            print('Building meta network')
        self.meta._build()
        if self.verbose:
            print('Tracing meta network')
        self.meta.trace(input_data, verbose=self.verbose)
        if self.verbose:
            print('Finalizing meta network')

        self.accumulate = accumulate
        self.num_last_filter_outputs = num_last_filter_outputs

        self.opt = opt
        self.opt_type = 'sgd'
        if isinstance(self.opt, AdamGnT) or 'Adam' in self.opt.__class__.__name__:
            self.opt_type = 'adam'

        """
        Define the hyper-parameters of the algorithm
        """
        self.replacement_rate = replacement_rate
        self.decay_rate = decay_rate
        self.maturity_threshold = maturity_threshold
        self.util_type = util_type

        """
        Utility of all features/neurons
        """
        self.util = {}
        self.bias_corrected_util = {}
        self.ages = {}
        self.mean_feature_act = {}
        self.mean_abs_feature_act = {}
        self.accumulated_num_features_to_replace = {}

        self.tracked_layer_names = []

        for name, layer in self.meta.named_layers:
            # FIXME: multiple inputs to the same layer may need to be handled
            # differently.

            if isinstance(layer, Conv2d):
                shape = layer.out_channels
                self.util[name] = torch.zeros(shape, device=device)
                self.bias_corrected_util[name] = torch.zeros(shape, device=device)
                self.ages[name] = torch.zeros(shape, device=device)
                self.mean_feature_act[name] = torch.zeros(shape, device=device)
                self.mean_abs_feature_act[name] = torch.zeros(shape, device=device)
                self.tracked_layer_names.append(name)
            elif isinstance(layer, Linear):
                shape = layer.out_features
                self.util[name] = torch.zeros(shape, device=device)
                self.bias_corrected_util[name] = torch.zeros(shape, device=device)
                self.ages[name] = torch.zeros(shape, device=device)
                self.mean_feature_act[name] = torch.zeros(shape, device=device)
                self.mean_abs_feature_act[name] = torch.zeros(shape, device=device)
                self.tracked_layer_names.append(name)
            self.accumulated_num_features_to_replace[name] = 0

        """
        Calculate uniform distribution's bound for random feature initialization
        """
        self.bounds = self.compute_bounds(
            hidden_activation=hidden_activation, init=init)

        """
        Pre calculate number of features to replace per layer per update
        """
        self.num_new_features_to_replace = {}
        for name in self.tracked_layer_names:
            layer = self.meta.name_to_layer[name]
            if isinstance(layer, Linear):
                self.num_new_features_to_replace[name] = (
                    self.replacement_rate * layer.out_features)
            elif isinstance(layer, Conv2d):
                self.num_new_features_to_replace[name] = (
                    self.replacement_rate * layer.out_channels)

    def compute_bounds(self, hidden_activation, init='kaiming'):
        if hidden_activation == 'selu':
            init = 'lecun'
        if hidden_activation in ['swish', 'elu']:
            hidden_activation = 'relu'
        bounds = {}
        gain = calculate_gain(nonlinearity=hidden_activation)
        for name in self.tracked_layer_names:
            layer = self.meta.name_to_layer[name]
            bound = get_layer_bound(layer=layer, init=init, gain=gain)
            bounds[name] = bound
        # # TODO: bounds seem to need before / after for each layer
        # bounds.append(get_layer_bound(layer=self.net[-1], init=init, gain=1))
        return bounds

    @staticmethod
    def _layer_type(layer):
        if isinstance(layer, Linear):
            typename = 'Linear'
        elif isinstance(layer, Conv2d):
            typename = 'Conv2d'
        else:
            raise AssertionError
        return typename

    def update_utility(self, layer_name, feature):
        with torch.no_grad():
            self.util[layer_name] *= self.decay_rate

            preserve_rate = (1 - self.decay_rate)
            bias_correction = 1 - (self.decay_rate ** self.ages[layer_name])

            curr_layer = self.meta.name_to_layer[layer_name]
            curr_type = self._layer_type(curr_layer)

            for next_layer_name in self.meta.next_layer_names(layer_name):
                # Not sure how to deal with multiple outputs here.
                next_layer = self.meta.name_to_layer[next_layer_name]

                next_type = self._layer_type(next_layer)
                try:
                    if next_type == 'Linear':
                        output_wight_mag = next_layer.weight.data.abs().mean(dim=0)
                    elif next_type == 'Conv2d':
                        output_wight_mag = next_layer.weight.data.abs().mean(dim=(0, 2, 3))
                        if next_layer.groups > 1:
                            output_wight_mag = output_wight_mag.repeat(next_layer.groups)
                    else:
                        raise NotImplementedError

                    self.mean_feature_act[layer_name] *= self.decay_rate
                    self.mean_abs_feature_act[layer_name] *= self.decay_rate

                    if curr_type == 'Linear':
                        input_wight_mag = curr_layer.weight.data.abs().mean(dim=1)
                    elif curr_type == 'Conv2d':
                        input_wight_mag = curr_layer.weight.data.abs().mean(dim=(1, 2, 3))
                        # if curr_layer.groups > 1:
                        #     input_wight_mag = input_wight_mag.repeat(curr_layer.groups)
                    else:
                        raise AssertionError

                    view_dims = None
                    if curr_type == 'Linear':
                        if feature.ndim == 2:
                            feat_mean_dim = 0
                        elif feature.ndim == 3:
                            feat_mean_dim = (0, 1)
                        elif feature.ndim == 4:
                            feat_mean_dim = (0, 1, 2)
                        else:
                            raise AssertionError(feature.shape)
                    elif curr_type == 'Conv2d' and next_type == 'Conv2d':
                        feat_mean_dim = (0, 2, 3)
                    elif curr_type == 'Conv2d' and next_type == 'Linear':
                        # FIXME: not general
                        feat_mean_dim = 0
                        view_dims = (-1, self.num_last_filter_outputs)
                    else:
                        raise AssertionError

                    act_update = -preserve_rate * feature.mean(dim=feat_mean_dim)
                    abs_act_update = -preserve_rate * feature.abs().mean(dim=feat_mean_dim)
                    if view_dims is not None:
                        act_update = act_update.view(*view_dims).mean(dim=1)
                        abs_act_update = abs_act_update.view(*view_dims).mean(dim=1)
                    self.mean_feature_act[layer_name] -= act_update
                    self.mean_abs_feature_act[layer_name] -= abs_act_update

                    bias_corrected_act = self.mean_feature_act[layer_name] / bias_correction

                    if self.util_type == 'adaptation':
                        new_util = 1 / input_wight_mag
                    elif self.util_type in ['contribution', 'zero_contribution', 'adaptable_contribution']:
                        if self.util_type == 'contribution':
                            bias_corrected_act = 0
                        else:
                            if curr_type == 'Conv2d' and next_type == 'Conv2d':
                                bias_corrected_act = bias_corrected_act.view(1, -1, 1, 1)
                            elif curr_type == 'Conv2d' and next_type == 'Linear':
                                bias_corrected_act = bias_corrected_act.repeat_interleave(
                                    self.num_last_filter_outputs).view(1, -1)

                        if curr_type == 'Linear' and next_type == 'Linear':
                            if feature.ndim == 2:
                                new_util = output_wight_mag * (feature - bias_corrected_act).abs().mean(dim=0)
                            elif feature.ndim == 3:
                                new_util = output_wight_mag * (feature - bias_corrected_act).abs().mean(dim=(0, 1))
                            elif feature.ndim == 4:
                                new_util = output_wight_mag * (feature - bias_corrected_act).abs().mean(dim=(0, 1, 2))
                            else:
                                raise AssertionError(feature.ndim)
                        elif curr_type == 'Conv2d' and next_type == 'Linear':
                            new_util = (output_wight_mag * (feature - bias_corrected_act).abs().mean(dim=0)).view(-1, self.num_last_filter_outputs).mean(dim=1)
                        elif next_type == 'Conv2d':
                            new_util = output_wight_mag * (feature - bias_corrected_act).abs().mean(dim=(0, 2, 3))

                        if self.util_type == 'adaptable_contribution':
                            new_util = new_util / input_wight_mag

                    if self.util_type == 'random':
                        self.bias_corrected_util[layer_name] = torch.rand(self.util[layer_name].shape)
                    else:
                        self.util[layer_name] -= -preserve_rate * new_util
                        # correct the bias in the utility computation
                        self.bias_corrected_util[layer_name] = self.util[layer_name] / bias_correction
                except Exception as ex:
                    e = ex
                    curr_nx_node = self.meta.name_to_nx_node[layer_name]
                    next_nx_node = self.meta.name_to_nx_node[next_layer_name]
                    curr_node_data = self.meta.nx_graph.nodes[curr_nx_node]
                    next_node_data = self.meta.nx_graph.nodes[next_nx_node]
                    print('curr_node_data = {}'.format(ub.urepr(curr_node_data, nl=1)))
                    print('next_node_data = {}'.format(ub.urepr(next_node_data, nl=1)))
                    # print(self.mean_feature_act[layer_name].shape)
                    rich.print(ub.codeblock(
                        f'''
                        [red]bad: {layer_name} -> {next_layer_name},
                        feature.shape={feature.shape}
                        {self.util[layer_name].shape=}
                        e={e!r}
                    '''))
                    import xdev
                    xdev.embed()
                    raise
                else:
                    ...
                    # rich.print(ub.codeblock(
                    #     f'''
                    #     [green]good: {layer_name} -> {next_layer_name},
                    #     feature.shape={feature.shape}
                    # '''))
                    # # raise

    def test_features(self, features):
        """
        Args:
            features: Activation values in the neural network
        Returns:
            Features to replace in each layer, Number of features to replace in each layer
        """
        features_to_replace = {n: torch.empty(0, dtype=torch.long, device=self.device)
                               for n in self.tracked_layer_names}
        features_to_replace_input_indices = {n: torch.empty(0, dtype=torch.long)
                                             for n in self.tracked_layer_names}
        features_to_replace_output_indices = {n: torch.empty(0, dtype=torch.long)
                                              for n in self.tracked_layer_names}
        num_features_to_replace = {n: 0 for n in self.tracked_layer_names}
        if self.replacement_rate == 0:
            return features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace

        for layer_name, feature in features.items():
            self.ages[layer_name] += 1
            """
            Update feature utility
            """
            self.update_utility(layer_name, feature)

            """
            Find the no. of features to replace
            """
            eligible_feature_indices = torch.where(self.ages[layer_name] > self.maturity_threshold)[0]
            if eligible_feature_indices.shape[0] == 0:
                continue
            new_num_replace = self.replacement_rate * eligible_feature_indices.shape[0]
            self.accumulated_num_features_to_replace[layer_name] += new_num_replace

            """
            Case when the number of features to be replaced is between 0 and 1.
            """
            if self.accumulate:
                new_num_replace = int(self.accumulated_num_features_to_replace[layer_name])
                self.accumulated_num_features_to_replace[layer_name] -= new_num_replace
            else:
                if new_num_replace < 1:
                    if torch.rand(1) <= new_num_replace:
                        new_num_replace = 1
                new_num_replace = int(new_num_replace)

            if new_num_replace == 0:
                continue

            """
            Find features to replace in the current layer
            """
            layer_bias_correction = self.bias_corrected_util[layer_name]
            _tmp = -layer_bias_correction[eligible_feature_indices]
            _new_replace_idxs = torch.topk(_tmp, new_num_replace)[1]
            new_replace_idxs = eligible_feature_indices[_new_replace_idxs]
            """
            Initialize utility for new features
            """
            self.util[layer_name][new_replace_idxs] = 0
            self.mean_feature_act[layer_name][new_replace_idxs] = 0.

            features_to_replace[layer_name] = new_replace_idxs
            num_features_to_replace[layer_name] = new_num_replace

        test_result = (features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace)
        return test_result

    def gen_new_features(self, test_result):
        """
        Generate new features: Reset input and output weights for low utility features
        """
        features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace = test_result
        with torch.no_grad():
            for layer_name in self.tracked_layer_names:

                num = num_features_to_replace[layer_name]

                if num == 0:
                    continue

                curr_layer = self.meta.name_to_layer[layer_name]
                in_feat_idx = features_to_replace_input_indices[layer_name]
                out_feat_idx = features_to_replace_output_indices[layer_name]
                bound = self.bounds[layer_name]

                if isinstance(curr_layer, Linear):
                    _shape = [num, curr_layer.in_features]
                elif isinstance(curr_layer, Conv2d):
                    _shape = [num] + list(curr_layer.weight.shape[1:])
                else:
                    raise AssertionError('should be linear or conv2d for now')
                curr_layer.weight.data[in_feat_idx, :] *= 0.0
                curr_layer.weight.data[in_feat_idx, :] -= -(
                    torch.empty(_shape, device=self.device).uniform_(-bound, bound))

                if curr_layer.bias is not None:
                    curr_layer.bias.data[in_feat_idx] *= 0.0

                # Set ages to zero
                self.ages[layer_name][in_feat_idx] = 0

                # Set the outgoing weights to zero
                for next_layer_name in self.meta.next_layer_names(layer_name):
                    next_layer = self.meta.name_to_layer[next_layer_name]
                    next_layer.weight.data[:, out_feat_idx] = 0

    def update_optim_params(self, test_result):
        """
        Update Optimizer's state
        """
        features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace = test_result
        if self.opt_type == 'adam':
            for layer_name in self.tracked_layer_names:

                # input weights
                if num_features_to_replace == 0:
                    continue

                layer = self.meta.name_to_layer[layer_name]

                for next_layer_name in self.meta.next_layer_names(layer_name):
                    next_layer = self.meta.name_to_layer[next_layer_name]

                    in_feat_idx = features_to_replace_input_indices[layer_name]
                    out_feat_idx = features_to_replace_output_indices[layer_name]

                    curr_weight_state = self.opt.state[layer.weight]
                    curr_bias_state = self.opt.state[layer.bias]
                    next_weight_state = self.opt.state[next_layer.weight]

                    if 'exp_avg' in curr_weight_state:
                        curr_weight_state['exp_avg'][in_feat_idx, :] = 0.0
                    if 'exp_avg_sq' in curr_weight_state:
                        curr_weight_state['exp_avg_sq'][in_feat_idx, :] = 0.0
                    if 'step' in curr_weight_state:
                        try:
                            curr_weight_state['step'][in_feat_idx] = 0
                        except Exception:
                            curr_weight_state['step'].zero_()

                    if 'exp_avg' in curr_bias_state:
                        curr_bias_state['exp_avg'][in_feat_idx] = 0.0
                    if 'exp_avg_sq' in curr_bias_state:
                        curr_bias_state['exp_avg_sq'][in_feat_idx] = 0.0
                    if 'step' in curr_bias_state:
                        try:
                            curr_bias_state['step'][in_feat_idx] = 0
                        except Exception:
                            curr_bias_state['step'].zero_()

                    if 'exp_avg' in next_weight_state:
                        next_weight_state['exp_avg'][:, out_feat_idx] = 0.0
                    if 'exp_avg_sq' in next_weight_state:
                        next_weight_state['exp_avg_sq'][:, out_feat_idx] = 0.0
                    if 'step' in next_weight_state:
                        try:
                            next_weight_state['step'][:, out_feat_idx] = 0
                        except Exception:
                            next_weight_state['step'].zero_()

    def current_features(self):
        running_features = self.meta.activation_cache
        running_features = {k: v for k, v in running_features.items() if k in self.tracked_layer_names}
        features = {k: v.mean() for k, v in running_features.items()}
        return features

    def clear_features(self):
        for v in self.meta.activation_cache.values():
            v.clear()

    def gen_and_test(self):
        """
        Perform generate-and-test
        """
        features = self.current_features()
        test_result = self.test_features(features=features)
        self.gen_new_features(test_result)
        self.update_optim_params(test_result)
        self.clear_features()


def model_layers(model):
    """
    Note:
        This was moved to netharn.initializers.functional.
        Move it back here, or do some other refactoring.

    Example:
        >>> import torchvision
        >>> model = torchvision.models.AlexNet()
        >>> list(model_layers(model))
    """
    stack = [('', '', model)]
    while stack:
        prefix, basename, item = stack.pop()
        name = '.'.join([p for p in [prefix, basename] if p])
        if isinstance(item, torch.nn.modules.conv._ConvNd):
            yield name, item
        elif isinstance(item, torch.nn.modules.batchnorm._BatchNorm):
            yield name, item
        elif hasattr(item, 'reset_parameters'):
            yield name, item

        child_prefix = name
        for child_basename, child_item in list(item.named_children())[::-1]:
            stack.append((child_prefix, child_basename, child_item))


class MetaNetwork:
    """
    Stores extra information that we need about the network

    Example:
        >>> import geowatch_tpl
        >>> geowatch_tpl.import_submodule('torchview')
        >>> geowatch_tpl.import_submodule('lop')
        >>> import torchvision
        >>> net = torchvision.models.resnet18()
        >>> meta = MetaNetwork(net)._build()
        >>> meta.trace(torch.rand(2, 3, 224, 224))
        >>> import networkx as nx
        >>> nx.write_network_text(meta.nx_graph, vertical_chains=1)
        >>> import torchvision
        >>> inputs = torch.rand(2, 3, 224, 224)
        >>> outputs = net(inputs)
        >>> act_shapes = ub.udict(meta.activation_cache).map_values(lambda x: x.shape)
        >>> print('act_shapes = {}'.format(ub.urepr(act_shapes, nl=1)))

    Ignore:
        traced = torch.jit.trace(net, inputs)
    """
    def __init__(meta, net):
        meta.net = net
        meta.activation_cache = {}
        meta.named_layers = None
        meta.id_to_names = None
        meta.name_to_layer = None

    def _build(meta):
        meta._build_layer_information()
        meta._register_layer_hooks()
        return meta

    def _build_layer_information(meta):
        net = meta.net
        named_layers = list(model_layers(net))
        name_to_layer = dict(named_layers)
        id_to_names = defaultdict(list)
        for name, layer in named_layers:
            layer_id = id(layer)
            id_to_names[layer_id].append(name)
        meta.id_to_names = id_to_names
        meta.named_layers = named_layers
        meta.name_to_layer = name_to_layer
        meta.layers = [t[1] for t in meta.named_layers]
        meta.layer_names = [t[0] for t in meta.named_layers]

        meta.type_to_layers = defaultdict(list)
        for _, layer in meta.named_layers:
            meta.type_to_layers[layer.__class__].append(layer)

    def _register_layer_hooks(meta):

        for name, layer in meta.named_layers:
            layer._forward_hooks.clear
            hook = RecordActivationHook(name, meta)
            layer.register_forward_hook(hook)

    def trace(meta, input_data=None, input_shape=None, verbose=1):
        """
        Requires an example input
        """
        # from torchview import draw_graph
        import networkx as nx
        import copy

        net = meta.net
        if verbose:
            print('deep copy network')
        net_copy = copy.deepcopy(net)
        if verbose:
            print('finish deep copy')

        id_to_names = defaultdict(list)
        for name, layer in list(model_layers(net_copy)):
            layer_id = id(layer)
            id_to_names[layer_id].append(name)

        tv_graph = patched_trace_graph(net_copy, input_data)

        def make_label(n_id, data):
            """ Create a nice printable label """
            n_id_str = str(n_id)
            parts = []
            if 'layer_name' in data:
                parts.append(data['layer_name'] + ':')
            if 'tv_compute_node' in data:
                n = data['tv_compute_node']
                parts.append(n.name)
            else:
                parts.append(n_id_str)
            if n_id_str in tv_graph.id_dict:
                idx = tv_graph.id_dict[n_id_str]
                parts.append(f':{idx}')

            if n_id in id_to_names:
                parts.append(' ' + id_to_names[n_id])

            label = ''.join(parts)
            return label

        # Build a networkx version of the torchview model nx_graph
        nx_graph = nx.DiGraph()
        for node in tv_graph.node_set:
            nx_graph.add_node(node)

        for tv_u, tv_v in tv_graph.edge_list:
            tv_u_id = id(tv_u)
            tv_v_id = id(tv_v)
            nx_graph.add_edge(tv_u_id, tv_v_id)
            nx_graph.nodes[tv_u_id]['tv_compute_node'] = tv_u
            nx_graph.nodes[tv_v_id]['tv_compute_node'] = tv_v

        name_to_nx_node = {}

        # Enrich each node with more info
        for nx_node, data in nx_graph.nodes(data=True):
            if 'tv_compute_node' in data:
                tv_node = data['tv_compute_node']
                if hasattr(tv_node, 'compute_unit_id'):
                    if tv_node.compute_unit_id in id_to_names:
                        layer_names = id_to_names[tv_node.compute_unit_id]
                        if len(layer_names) == 1:
                            layer_name = data['layer_name'] = layer_names[0]
                            name_to_nx_node[layer_name] = nx_node
                        else:
                            data['layer_names'] = layer_names[0]
            data['label'] = make_label(nx_node, data)

        # Not sure what the rando singleton node is.
        if len(nx_graph.nodes) > 1:
            singleton_nodes = []
            for nx_node in nx_graph.nodes:
                if nx_graph.in_degree[nx_node] == 0 and nx_graph.out_degree[nx_node] == 0:
                    singleton_nodes.append(nx_node)
            nx_graph.remove_nodes_from(singleton_nodes)

        meta.nx_graph = nx_graph
        # nx.write_network_text(nx_graph, vertical_chains=1)

        # Determine which nodes have associated layer names
        named_ids = []
        for n_id, data in nx_graph.nodes(data=True):
            if 'layer_name' in data:
                named_ids.append(n_id)

        import ubelt as ub
        topo_order = ub.OrderedSet(nx.topological_sort(nx_graph))
        meta.named_topo_order = (topo_order & named_ids)
        meta.name_to_nx_node = name_to_nx_node
        meta.tv_graph = tv_graph

        # import netharn as nh

        # Build fast lookup for next layers
        meta.layer_dependencies = ub.ddict(list)
        for layer_name in meta.layer_names:
            next_names = list([t[0] for t in meta.next_layers(layer_name)])
            meta.layer_dependencies[layer_name].extend(next_names)
            if layer_name in meta.name_to_nx_node:
                nx_node = meta.name_to_nx_node[layer_name]
                node_data = meta.nx_graph.nodes[nx_node]
                if 'tv_compute_node' in node_data:
                    tv_node = node_data['tv_compute_node']
                    out_shape = tv_node.output_shape
                    in_shape = tv_node.input_shape
                    node_data['in_shape'] = in_shape
                    node_data['out_shape'] = out_shape
                node_data['next_names'] = next_names

        # tv_graph.visual_graph.view()

    def next_layer_names(meta, layer_name):
        return meta.layer_dependencies[layer_name]

    def next_layers(meta, layer_name):
        """
        import netharn as nh
        import rich
        for layer_name in meta.layer_names:
            rich.print('[blue] layer_name = {}'.format(ub.urepr(layer_name, nl=1)))
            children = list([t[0] for t in meta.next_layers(layer_name)])
            print('children = {}'.format(ub.urepr(children, nl=1)))
        """
        import networkx as nx
        curr_name = layer_name
        if curr_name in meta.name_to_nx_node:
            curr_nx_node = meta.name_to_nx_node[curr_name]

            curr_node_data = meta.nx_graph.nodes[curr_nx_node]
            if 'tv_compute_node' in curr_node_data:
                curr_tv_node = curr_node_data['tv_compute_node']
                # curr_in_shape = n.input_shape
                curr_out_shape = curr_tv_node.output_shape

            named_descendants = meta.named_topo_order & nx.descendants(meta.nx_graph, curr_nx_node)
            for next_nx_node in named_descendants:
                next_layer_name = meta.nx_graph.nodes[next_nx_node]['layer_name']
                next_layer = meta.name_to_layer[next_layer_name]
                flag = 0
                if isinstance(next_layer, KNOWN_LAYERS):
                    next_node_data = meta.nx_graph.nodes[next_nx_node]
                    if 'tv_compute_node' in next_node_data:
                        path = nx.shortest_path(meta.nx_graph, curr_nx_node, next_nx_node)
                        next_tv_node = next_node_data['tv_compute_node']
                        next_in_shape = next_tv_node.input_shape
                        between = path[1:-1]
                        between_datas = [meta.nx_graph.nodes[_n] for _n in between]
                        between_named = [d.get('layer_name', None) for d in between_datas]
                        between_layers = [meta.name_to_layer[n] for n in between_named if n is not None]
                        n_between = len([layer for layer in between_layers if isinstance(layer, KNOWN_LAYERS)])
                        # between_flags = sum([b is None for b in between_named])
                        # meta.name_to_layer
                        # next_out_shape = next_tv_node.output_shape
                        if n_between == 0:
                            if (next_in_shape) == curr_out_shape:
                                flag = 1
                    if flag:
                        # print(f'{curr_name} -> {next_layer_name}')
                        # print(between_named)
                        # # named_path_datas = [meta.nx_graph.nodes[_n] for _n in path]
                        # # path_labels = [d['label'] for d in named_path_datas]
                        # # print(path_labels)
                        # print(f'{curr_out_shape} -> {next_in_shape}')
                        yield next_layer_name, next_layer


class TorchRunningStats:
    """
    Example:
        import torch
        data = torch.rand(1, 3, 32, 32, device=0)
        self = TorchRunningStats()
        self.update(data)
        self.update(torch.rand(1, 3, 32, 32, device=0))
        self.update(torch.rand(1, 3, 32, 32, device=0))
        self.update(torch.rand(1, 3, 32, 32, device=0))
        self.summarize()
    """
    def __init__(self, device=None):
        self.n = 0
        self.raw_total = 0
        self.device = device

    def clear(self):
        self.n = 0
        self.raw_total = 0

    def update(self, data):
        self.raw_total = self.raw_total + data
        self.n += 1

    @property
    def shape(self):
        return self.raw_total.shape

    def mean(self):
        n = self.n
        total = self.raw_total
        if n == 1:
            return total
        return total / n

    def summarize(self):
        n = self.n
        total = self.raw_total
        info = {
            'total': total,
            'mean': total / n,
            'n': n,
        }
        return info


class RecordActivationHook:
    def __init__(self, name, meta):
        self.name = name
        self.meta = meta

    def __call__(self, layer, input, output):
        activation = output.detach()
        if self.name not in self.meta.activation_cache:
            self.meta.activation_cache[self.name] = TorchRunningStats(device=activation.device)

        try:
            runner = self.meta.activation_cache[self.name]
            runner.raw_total = activation
            runner.n = 1
            # runner.update(activation)
        except Exception:
            return
            from watch.tasks.fusion.methods.channelwise_transformer import slice_to_agree
            # Could do this as a hack...
            a, b = slice_to_agree(activation, runner.raw_total)
            # print(f'a.shape={a.shape}')
            # print(f'b.shape={b.shape}')
            # print(f'activation.shape       = {activation.shape}')
            # print(f'runner.raw_total.shape = {runner.raw_total.shape}')

            if np.prod(a.shape) > 0:
                activation = a
                runner.raw_total = b
                runner.update(activation)


def patched_trace_graph(net_copy, input_data):
    """
    References:
        https://discuss.pytorch.org/t/tracing-a-graph-of-torch-layers/187615/2
        https://github.com/mert-kurttutan/torchview/issues/104
        https://github.com/mert-kurttutan/torchview/issues/103
        https://github.com/mert-kurttutan/torchview/pull/105
    """
    input_size = None

    if 1:
        ### monkeypatch torchview
        from torchview import torchview as torchview_submod
        if not getattr(torchview_submod, '_geowatch_monkey_patched', 0):
            torchview_submod._geowatch_monkey_patched = 1
            _orig_traverse_data = getattr(
                torchview_submod, '_orig_traverse_data',
                torchview_submod.traverse_data)
            torchview_submod._orig_traverse_data = _orig_traverse_data
            assert torchview_submod.traverse_data.__name__ != 'new_traverse_data'
            def new_traverse_data(data, action_fn, aggregate_fn):
                if isinstance(data, np.ndarray):
                    return data
                return _orig_traverse_data(data, action_fn, aggregate_fn)
            torchview_submod.traverse_data = new_traverse_data
    if 1:
        depth = 9001
        dtypes = None
        graph_name = 'model'
        filename = f'{graph_name}.gv'
        # device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        device = torch.device("cpu")
        # if torch.cuda.is_available() else "cpu")
        model_mode = 'eval'
        graph_dir = 'TB'
        from torchview.torchview import validate_user_params
        from torchview.torchview import ComputationGraph
        from torchview.torchview import process_input
        from torchview.torchview import module_forward_wrapper
        from torchview.torchview import _orig_module_forward
        from torchview.torchview import Recorder
        import graphviz
        from typing import Mapping
        model = net_copy
        validate_user_params(
            model, input_data, input_size, depth, device, dtypes,
        )
        graph_attr = {
            'ordering': 'in',
            'rankdir': graph_dir,
        }

        # visual settings from torchviz
        # seems to work visually well
        node_attr = {
            'style': 'filled',
            'shape': 'plaintext',
            'align': 'left',
            'fontsize': '10',
            'ranksep': '0.1',
            'height': '0.2',
            'fontname': 'Linux libertine',
            'margin': '0',
        }

        edge_attr = {
            'fontsize': '10',
        }
        visual_graph = graphviz.Digraph(
            name=graph_name, engine='dot', strict=True,
            graph_attr=graph_attr, node_attr=node_attr, edge_attr=edge_attr,
            directory='.', filename=filename
        )

        input_recorder_tensor, kwargs_record_tensor, input_nodes = process_input(
            input_data, input_size, {}, device, dtypes
        )

        tv_graph = ComputationGraph(
            visual_graph, input_nodes, show_shapes=1, expand_nested=1,
            hide_inner_tensors=1, hide_module_functions=1, roll=0,
            depth=depth
        )

        # forward_prop(
        #     model, input_recorder_tensor, device, tv_graph,
        #     model_mode, **kwargs_record_tensor
        # )
        saved_model_mode = model.training
        try:
            mode = model_mode
            if mode == 'train':
                model.train()
            elif mode == 'eval':
                model.eval()
            else:
                raise RuntimeError(
                    f"Specified model mode not recognized: {mode}"
                )
            x = input_recorder_tensor
            new_module_forward = module_forward_wrapper(tv_graph)
            recorder = Recorder(_orig_module_forward, new_module_forward, tv_graph)
            with recorder:
                with torch.no_grad():
                    try:
                        kwargs = {}
                        if isinstance(x, (list, tuple)):
                            _ = model.to(device)(*x, **kwargs)
                        elif isinstance(x, Mapping):
                            _ = model.to(device)(**x, **kwargs)
                    except Exception:
                        _ = model.to(device)(x)
        except Exception as e:
            raise RuntimeError(
                "Failed to run torchgraph see error message"
            ) from e
        finally:
            model.train(saved_model_mode)

        tv_graph.fill_visual_graph()
        # tv_graph.visual_graph.render(format='png')
        # tv_graph.visual_graph.view()
    else:
        from torchview import draw_graph
        tv_graph = draw_graph(
            net_copy,
            input_data=input_data,
            input_size=input_size,
            expand_nested=True,
            hide_inner_tensors=True,
            device='meta', depth=np.inf)
        # tv_graph.visual_graph.view()
    return tv_graph



import torch


def get_real(t, complex_dim=-1):
    return t.select(complex_dim, 0)


def get_imag(t, complex_dim=-1):
    return t.select(complex_dim, 1)


def complex_mul(t1, t2, complex_dim=-1):
    t1_real = get_real(t1, complex_dim)
    t1_imag = get_imag(t1, complex_dim)
    t2_real = get_real(t2, complex_dim)
    t2_imag = get_imag(t2, complex_dim)

    ac = t1_real * t2_real
    bd = t1_imag * t2_imag
    ad = t1_real * t2_imag
    bc = t1_imag * t2_real
    tr_real = ac - bd
    tr_imag = ad + bc

    tr = torch.stack([tr_real, tr_imag], dim=complex_dim)

    return tr


def complex_sqrt(t, complex_dim=-1):
    sqrt_t_abs = torch.sqrt(complex_abs(t, complex_dim))
    sqrt_t_arg = complex_arg(t, complex_dim) / 2
    # Overwrite t with cos(\theta / 2) + i sin(\theta / 2):
    sqrt_t = sqrt_t_abs.unsqueeze(complex_dim) * torch.stack([torch.cos(sqrt_t_arg), torch.sin(sqrt_t_arg)], dim=complex_dim)
    return sqrt_t


def complex_abs_squared(t, complex_dim=-1):
    return get_real(t, complex_dim)**2 + get_imag(t, complex_dim)**2


def complex_abs(t, complex_dim=-1):
    return torch.sqrt(complex_abs_squared(t, complex_dim=complex_dim))


def complex_arg(t, complex_dim=-1):
    return torch.atan2(get_imag(t, complex_dim), get_real(t, complex_dim))


def main():
    device = None

    t1 = torch.Tensor([
        [2, 0],
        [0, 2],
        [-1, 0],
        [0, -1],
    ]).to(device)
    t2 = torch.Tensor([
        [2, 0],
        [0, 2],
        [-1, 0],
        [0, -1],
    ]).to(device)
    complex_dim = -1

    print(t1.int())
    print(t2.int())
    t1_mul_t2 = complex_mul(t1, t2, complex_dim)
    print(t1_mul_t2.int())

    sqrt_t1_mul_t2 = complex_sqrt(t1_mul_t2)
    print(sqrt_t1_mul_t2.int())


if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Drop path operation.

This scripts implements the drop path operation, proposed in
Gao Huang, Yu Sun, Zhuang Liu, Daniel Sedra, Kilian Weinberger,
Deep Networks with Stochastic Depth. In ECCV, 2016.
"""
import tensorflow as tf
from deeplab2.model import utils


def get_drop_path_keep_prob(keep_prob_for_last_stage, schedule,
                            current_stage, num_stages):
    """Gets drop path keep probability for current stage.

    Args:
      keep_prob_for_last_stage: A float, the drop path keep probability for
        last stage. This flag is used in conjunction with the flag `schedule`, as
        they together determine drop path keep probability for the other stages.
      schedule: A string, the drop path schedule. Currently, we support
        'constant': use the same drop path keep probability for all stages, and
        'linear': linearly decrease the drop path keep probability from 1.0 at
          0-th stage (or STEM) to `keep_prob_for_last_stage` at last stage.
      current_stage:  An integer, current stage number.
      num_stages: An integer, the number of stages.

    Returns:
      The drop path keep probability for the current stage.

    Raises:
      ValueError: If schedule is not supported.
    """
    if schedule == 'constant':
        return keep_prob_for_last_stage
    elif schedule == 'linear':
        return 1.0 - (1.0 - keep_prob_for_last_stage) * current_stage / num_stages
    else:
        raise ValueError('Unexpected schedule %s.' % schedule)


def generate_drop_path_random_mask(input_tensor, drop_path_keep_prob):
    """Generates a random mask for drop path.

    This function generates a random mask for training models with drop path. Each
    scalar in the output indicates whether the block or path will be kept. The
    scalars are scaled with (1.0 / drop_path_keep_prob) so that the output will
    have the same expectation no mather what the drop_path_keep_prob is.

    Reference:
    "Deep Networks with Stochastic Depth" https://arxiv.org/pdf/1603.09382.pdf

    Args:
      input_tensor: An input [batch_size, n_1, n_2, ..., n_k] tensor.
      drop_path_keep_prob: A float, the keep probability for dropping path.

    Returns:
      binary_tensor: A [batch_size, 1, 1, ..., 1] tensor with the same dtype as
        the input_tensor.
    """
    binary_tensor = None
    if drop_path_keep_prob < 1.0:
        input_shape = input_tensor.get_shape().as_list()
        batch_size = utils.resolve_batch_size(input_tensor)
        random_tensor_shape = [batch_size] + [1] * (len(input_shape) - 1)
        random_tensor = drop_path_keep_prob
        random_tensor += tf.random.uniform(
            random_tensor_shape, dtype=input_tensor.dtype)
        binary_tensor = tf.math.divide(tf.floor(random_tensor), drop_path_keep_prob)
    return binary_tensor


class DropPath(tf.keras.layers.Layer):
    """Drop path layer.

    For details, please see the original paper listed below.
    Gao Huang, Yu Sun, Zhuang Liu, Daniel Sedra, Kilian Weinberger,
    Deep Networks with Stochastic Depth. In ECCV, 2016.
    """

    def __init__(self, drop_path_keep_prob=1.0, name=None):
        """Initializes a drop path layer.

        Args:
          drop_path_keep_prob: A float, the keep probability for dropping path.
          name: An optional string specifying the operation name.

        Rasies:
          ValueError: If drop_path_keep_prob is <= 0 or > 1.
        """
        super(DropPath, self).__init__(name=name)
        self._drop_path_keep_prob = drop_path_keep_prob
        if self._drop_path_keep_prob <= 0 or self._drop_path_keep_prob > 1.0:
            raise ValueError('drop_path_keep_prob not valid. Got %f.' %
                             self._drop_path_keep_prob)

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        if training:
            keep_prob = self._drop_path_keep_prob
            shape = (tf.shape(input_tensor)[0],) + (1,) * (
                    len(tf.shape(input_tensor)) - 1)
            random_tensor = keep_prob + tf.random.uniform(shape, 0, 1)
            random_tensor = tf.floor(random_tensor)
            return (input_tensor / keep_prob) * random_tensor
        return input_tensor

    def get_config(self):
        config = {
            'drop_path_keep_prob': self._drop_path_keep_prob,
        }
        base_config = super(DropPath, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



#!/usr/bin/env python3
"""
THIS IS THE OLD FIT SCRIPT. MOST OF IT HAS BEEN REMOVED.
USE fit_lightning.py INSTEAD.

Trains a fusion machine learning model on target dataset.

SeeAlso:
    README.md
    fit.py
    predict.py
    evaluate.py
    experiments/crall/onera_experiments.sh
    experiments/crall/drop1_experiments.sh
    experiments/crall/toy_experiments.sh
"""


def coerce_initializer(init):
    import os
    import ubelt as ub
    from geowatch.monkey import monkey_torchmetrics
    from geowatch.monkey import monkey_torch
    monkey_torchmetrics.fix_torchmetrics_compatability()

    initializer = None

    maybe_packaged_model = False
    if isinstance(init, (str, os.PathLike)):
        if ub.Path(init).exists():
            maybe_packaged_model = True

    if maybe_packaged_model:
        try:
            from geowatch.tasks.fusion import utils
            other_model = utils.load_model_from_package(init)
            monkey_torch.fix_gelu_issue(other_model)
        except Exception:
            print('Not a packaged model')
        else:
            from torch_liberator.initializer import Pretrained
            import torch
            import tempfile
            tfile = tempfile.NamedTemporaryFile(prefix='pretrained_state', suffix='.pt')
            # state_dict = other_model.state_dict()
            try:
                state_dict = other_model.state_dict()
            except Exception:
                if hasattr(other_model, 'head_metrics'):
                    other_model.head_metrics.clear()
                    state_dict = other_model.state_dict()
                else:
                    raise

            # HACK:
            # Remove the normalization keys, we don't want to transfer them
            # in this step. They will be set correctly depending on if
            # normalize_inputs=transfer or not.
            HACK_IGNORE_INPUT_NORMS = True
            if HACK_IGNORE_INPUT_NORMS:
                ignore_keys = [key for key in state_dict if 'input_norms' in key]
                for k in ignore_keys:
                    state_dict.pop(k)
                print('Hacking a packaged model for init')

            # print(ub.urepr(sorted(state_dict.keys())))
            weights_fpath = tfile.name
            torch.save(state_dict, weights_fpath)
            init_cls = Pretrained
            init_kw = {'fpath': tfile.name}
            initializer = init_cls(**init_kw)
            # keep the temporary file alive as long as the initializer is
            initializer._tfile = tfile
            initializer.other_model = other_model

    if initializer is None:
        # Try a netharn method (todo: port to geowatch to remove netharn deps)
        from geowatch.utils import util_netharn
        init_cls, init_kw = util_netharn.Initializer.coerce(init=init)
        initializer = init_cls(**init_kw)

    return initializer


"""
Ignore:
"""



#!/usr/bin/env python3
"""
See Old Version:
    ../../../scripts/run_sc_fusion_eval3_for_baseline.py

SeeAlso:
    ~/code/watch-smartflow-dags/KIT_TA2_PREEVAL10_PYENV.py
"""
import sys
import traceback
import shutil

import scriptconfig as scfg
import ubelt as ub
from geowatch.mlops.smart_pipeline import DinoBoxDetector, SV_DinoFilter


class DinoSVConfig(scfg.DataConfig):
    """
    Run TA-2 SC fusion as baseline framework component
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))
    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    aws_profile = scfg.Value(None, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))
    dino_detect_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Raw json/yaml or a path to a json/yaml file that specifies the
        config for DinoBoxDetector.
        '''))
    dino_filter_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Raw json/yaml or a path to a json/yaml file that specifies the
        config for SV_DinoFilter.
        '''))
    skip_on_fail = scfg.Value(False, help=ub.paragraph(
        '''
        If an error occurs, pass through input region / sites unchanged.
        '''))

    input_region_models_asset_name = scfg.Value('depth_filtered_regions', type=str, required=False, help=ub.paragraph(
        '''
        Which region model assets to use as input
        '''))

    input_site_models_asset_name = scfg.Value('depth_filtered_sites', type=str, required=False, help=ub.paragraph(
        '''
        Which site model assets to to use as input
        '''))


def main():
    config = DinoSVConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_dino_sv(config)


def run_dino_sv(config):
    from geowatch.cli.smartflow_ingress import smartflow_ingress
    from geowatch.cli.smartflow_egress import smartflow_egress
    from geowatch.utils.util_framework import download_region, determine_region_id
    from kwutil.util_yaml import Yaml
    from geowatch.utils import util_framework

    input_path = config.input_path
    input_region_path = config.input_region_path
    output_path = config.output_path

    outbucket = config.outbucket
    aws_profile = config.aws_profile
    dryrun = config.dryrun

    ####
    # DEBUGGING:
    # Print info about what version of the code we are running on
    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = ub.Path('/tmp/ingress')
    ingressed_assets = smartflow_ingress(
        input_path=input_path,
        assets=[
            'cropped_kwcoco_for_sv',
            'cropped_kwcoco_for_sv_assets',
            config.input_region_models_asset_name,
            config.input_site_models_asset_name,
        ],
        outdir=ingress_dir,
        aws_profile=aws_profile,
        dryrun=dryrun
    )

    # # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'

    download_region(
        input_region_path=input_region_path,
        output_region_path=local_region_path,
        aws_profile=aws_profile,
        strip_nonregions=True,
    )

    # Determine the region_id in the region file.
    region_id = determine_region_id(local_region_path)
    print(f'region_id={region_id}')

    dino_boxes_kwcoco_path = ingress_dir / 'dino_boxes_kwcoco.json'

    input_region_dpath = ub.Path(ingressed_assets[config.input_region_models_asset_name])
    input_sites_dpath = ub.Path(ingressed_assets[config.input_site_models_asset_name])

    input_region_fpath = ub.Path(input_region_dpath) / f'{region_id}.geojson'
    assert input_region_fpath.exists()

    # NOTE; we want to be using the output of SV crop, not necesarilly the the
    # dzyne output referenced by ingress_kwcoco_path
    # input_kwcoco_fpath = ingress_kwcoco_path
    input_kwcoco_fpath = ingressed_assets['cropped_kwcoco_for_sv']

    # FIXME: these output directories for region / site models should be passed
    # to us from the DAG
    output_sites_dpath = ingress_dir / 'sv_out_site_models'
    output_region_dpath = ingress_dir / 'sv_out_region_models'
    output_site_manifest_dpath = ingress_dir / 'tracking_manifests_sv'
    output_region_fpath = output_region_dpath / f'{region_id}.geojson'
    output_site_manifest_fpath = output_site_manifest_dpath / 'site_models_manifest.json'

    node_state.print_current_state(ingress_dir)

    # 3.1. Check that we have at least one "video" (BAS identified
    # site) to run over; if not skip SV fusion and KWCOCO to GeoJSON
    import kwcoco
    input_coco_dset = kwcoco.CocoDataset(input_kwcoco_fpath)
    print('input_coco_dset = {}'.format(ub.urepr(input_coco_dset, nl=1)))
    num_videos = input_coco_dset.n_videos
    # Note: cant open with json here because kwcoco will save compressed files
    # with open(input_kwcoco_fpath) as f:
    #     ingress_kwcoco_data = json.load(f)
    # num_videos = len(ingress_kwcoco_data.get('videos', ()))
    print(f'num_videos={num_videos}')
    output_region_dpath.ensuredir()

    if num_videos == 0:
        # Copy input region model into region_models outdir to be updated
        # (rather than generated from tracking, which may not have the
        # same bounds as the original)

        # Not sure if the above case is the right comment, but leaving this
        # here to guarentee the region with site summaries is passed forward
        # TODO: the dino code should just be robust to this.
        input_sites_dpath.copy(output_sites_dpath)
        input_region_fpath.copy(output_region_fpath)
    else:
        output_site_manifest_dpath.ensuredir()
        output_sites_dpath.ensuredir()
        # 3.2 Run DinoBoxDetector
        print("* Running Dino Detect *")

        default_dino_detect_config = ub.udict({
            'coco_fpath': input_kwcoco_fpath,
            'package_fpath': None,
            'batch_size': 1,
            'device': 0})
        dino_detect_config = (default_dino_detect_config
                              | Yaml.coerce(config.dino_detect_config or {}))

        if dino_detect_config.get('package_fpath', None) is None:
            raise ValueError('Requires package_fpath')

        dino_box_detector = DinoBoxDetector(root_dpath='/tmp/ingress')
        dino_box_detector.configure({
            'out_coco_fpath': dino_boxes_kwcoco_path,
            **dino_detect_config})

        try:
            ub.cmd(dino_box_detector.command(), check=True, verbose=3, system=True)

            # 3.3 Run SV_DinoFilter
            print("* Running Dino Building Filter *")

            default_dino_filter_config = ub.udict({})
            dino_filter_config = (default_dino_filter_config
                                  | Yaml.coerce(config.dino_filter_config or {}))

            dino_building_filter = SV_DinoFilter(root_dpath='/tmp/ingress')

            dino_building_filter.configure({
                'input_kwcoco': dino_boxes_kwcoco_path,
                'input_region': input_region_fpath,
                'input_sites': input_sites_dpath,
                'output_region_fpath': output_region_fpath,
                'output_sites_dpath': output_sites_dpath,
                'output_site_manifest_fpath': output_site_manifest_fpath,
                **dino_filter_config,
            })

            ub.cmd(dino_building_filter.command(), check=True, verbose=3, system=True)
        except Exception:
            if config.skip_on_fail:
                print("WARNING: Exception occurred (printed below), passing input sites / region models as output")
                traceback.print_exception(*sys.exc_info())

                shutil.copytree(input_sites_dpath, output_sites_dpath, dirs_exist_ok=True)
                shutil.copytree(input_region_dpath, output_region_dpath, dirs_exist_ok=True)
            else:
                raise
        else:
            # Validate and fix all outputs
            util_framework.fixup_and_validate_site_and_region_models(
                region_dpath=output_region_fpath.parent,
                site_dpath=output_sites_dpath,

            )

    node_state.print_current_state(ingress_dir)

    # 5. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    ingressed_assets['sv_out_site_models'] = output_sites_dpath
    ingressed_assets['sv_out_region_models'] = output_region_dpath
    if dino_boxes_kwcoco_path.exists():
        # Reroot kwcoco files to make downloaded results easier to work with
        ub.cmd(['kwcoco', 'reroot', f'--src={dino_boxes_kwcoco_path}', '--inplace=1', '--absolute=0'])
        ingressed_assets['sv_dino_boxes_kwcoco'] = dino_boxes_kwcoco_path

    smartflow_egress(ingressed_assets,
                     local_region_path,
                     output_path,
                     outbucket,
                     aws_profile=None,
                     dryrun=False,
                     newline=False)


if __name__ == "__main__":
    main()
